from __future__ import annotations

import collections
import typing
import warnings
from collections.abc import AsyncIterator, Mapping
from typing import TYPE_CHECKING, Any, Literal, TypeAlias, cast, overload

from cognite.client._api.sequence_data import SequencesDataAPI
from cognite.client._api_client import APIClient
from cognite.client._constants import DEFAULT_LIMIT_READ
from cognite.client.data_classes import (
    Sequence,
    SequenceFilter,
    SequenceList,
    SequenceUpdate,
    filters,
)
from cognite.client.data_classes._base import CogniteResource, PropertySpec
from cognite.client.data_classes.aggregations import AggregationFilter, UniqueResultList
from cognite.client.data_classes.filters import _BASIC_FILTERS, Filter, _validate_filter
from cognite.client.data_classes.sequences import (
    SequenceColumnUpdate,
    SequenceColumnWriteList,
    SequenceCore,
    SequenceProperty,
    SequenceSort,
    SequenceWrite,
    SortableSequenceProperty,
)
from cognite.client.data_classes.shared import TimestampRange
from cognite.client.utils._identifier import IdentifierSequence
from cognite.client.utils._validation import (
    assert_type,
    prepare_filter_sort,
    process_asset_subtree_ids,
    process_data_set_ids,
)
from cognite.client.utils.useful_types import SequenceNotStr

if TYPE_CHECKING:
    from cognite.client import AsyncCogniteClient
    from cognite.client.config import ClientConfig

SortSpec: TypeAlias = (
    SequenceSort
    | str
    | SortableSequenceProperty
    | tuple[str, Literal["asc", "desc"]]
    | tuple[str, Literal["asc", "desc"], Literal["auto", "first", "last"]]
)

_FILTERS_SUPPORTED: frozenset[type[Filter]] = _BASIC_FILTERS | {filters.Search}


class SequencesAPI(APIClient):
    _RESOURCE_PATH = "/sequences"

    def __init__(self, config: ClientConfig, api_version: str | None, cognite_client: AsyncCogniteClient) -> None:
        super().__init__(config, api_version, cognite_client)
        self.data = SequencesDataAPI(config, api_version, cognite_client)

    @property
    def rows(self) -> SequencesDataAPI:
        warnings.warn(
            "The 'sequences.rows' property is deprecated and may be removed in the future. "
            "Use 'client.sequences.data' to future-proof your code.",
            DeprecationWarning,
        )
        return self.data

    @overload
    def __call__(
        self,
        chunk_size: None = None,
        name: str | None = None,
        external_id_prefix: str | None = None,
        metadata: dict[str, str] | None = None,
        asset_ids: typing.Sequence[int] | None = None,
        asset_subtree_ids: int | typing.Sequence[int] | None = None,
        asset_subtree_external_ids: str | SequenceNotStr[str] | None = None,
        data_set_ids: int | typing.Sequence[int] | None = None,
        data_set_external_ids: str | SequenceNotStr[str] | None = None,
        created_time: dict[str, Any] | None = None,
        last_updated_time: dict[str, Any] | None = None,
        limit: int | None = None,
        advanced_filter: Filter | dict[str, Any] | None = None,
        sort: SortSpec | list[SortSpec] | None = None,
    ) -> AsyncIterator[Sequence]: ...

    @overload
    def __call__(
        self,
        chunk_size: int,
        name: str | None = None,
        external_id_prefix: str | None = None,
        metadata: dict[str, str] | None = None,
        asset_ids: typing.Sequence[int] | None = None,
        asset_subtree_ids: int | typing.Sequence[int] | None = None,
        asset_subtree_external_ids: str | SequenceNotStr[str] | None = None,
        data_set_ids: int | typing.Sequence[int] | None = None,
        data_set_external_ids: str | SequenceNotStr[str] | None = None,
        created_time: dict[str, Any] | None = None,
        last_updated_time: dict[str, Any] | None = None,
        limit: int | None = None,
        advanced_filter: Filter | dict[str, Any] | None = None,
        sort: SortSpec | list[SortSpec] | None = None,
    ) -> AsyncIterator[SequenceList]: ...

    async def __call__(
        self,
        chunk_size: int | None = None,
        name: str | None = None,
        external_id_prefix: str | None = None,
        metadata: dict[str, str] | None = None,
        asset_ids: typing.Sequence[int] | None = None,
        asset_subtree_ids: int | typing.Sequence[int] | None = None,
        asset_subtree_external_ids: str | SequenceNotStr[str] | None = None,
        data_set_ids: int | typing.Sequence[int] | None = None,
        data_set_external_ids: str | SequenceNotStr[str] | None = None,
        created_time: dict[str, Any] | None = None,
        last_updated_time: dict[str, Any] | None = None,
        limit: int | None = None,
        advanced_filter: Filter | dict[str, Any] | None = None,
        sort: SortSpec | list[SortSpec] | None = None,
    ) -> AsyncIterator[Sequence] | AsyncIterator[SequenceList]:
        """Iterate over sequences

        Fetches sequences as they are iterated over, so you keep a limited number of objects in memory.

        Args:
            chunk_size (int | None): Number of sequences to return in each chunk. Defaults to yielding one event a time.
            name (str | None): Filter out sequences that do not have this *exact* name.
            external_id_prefix (str | None): Filter out sequences that do not have this string as the start of the externalId
            metadata (dict[str, str] | None): Filter out sequences that do not match these metadata fields and values (case-sensitive). Format is {"key1":"value1","key2":"value2"}.
            asset_ids (typing.Sequence[int] | None): Filter out sequences that are not linked to any of these assets.
            asset_subtree_ids (int | typing.Sequence[int] | None): Only include sequences that have a related asset in a subtree rooted at any of these assetIds. If the total size of the given subtrees exceeds 100,000 assets, an error will be returned.
            asset_subtree_external_ids (str | SequenceNotStr[str] | None): Only include sequences that have a related asset in a subtree rooted at any of these assetExternalIds. If the total size of the given subtrees exceeds 100,000 assets, an error will be returned.
            data_set_ids (int | typing.Sequence[int] | None): Return only sequences in the specified data set(s) with this id / these ids.
            data_set_external_ids (str | SequenceNotStr[str] | None): Return only sequences in the specified data set(s) with this external id / these external ids.
            created_time (dict[str, Any] | None):  Range between two timestamps. Possible keys are `min` and `max`, with values given as time stamps in ms.
            last_updated_time (dict[str, Any] | None):  Range between two timestamps. Possible keys are `min` and `max`, with values given as time stamps in ms.
            limit (int | None): Max number of sequences to return. Defaults to return all items.
            advanced_filter (Filter | dict[str, Any] | None): Advanced filter query using the filter DSL (Domain Specific Language). It allows defining complex filtering expressions that combine simple operations, such as equals, prefix, exists, etc., using boolean operators and, or, and not.
            sort (SortSpec | list[SortSpec] | None): The criteria to sort by. Defaults to desc for `_score_` and asc for all other properties. Sort is not allowed if `partitions` is used.

        Yields:
            Sequence | SequenceList: yields Sequence one by one if chunk_size is not specified, else SequenceList objects.
        """  # noqa: DOC404
        asset_subtree_ids_processed = process_asset_subtree_ids(asset_subtree_ids, asset_subtree_external_ids)
        data_set_ids_processed = process_data_set_ids(data_set_ids, data_set_external_ids)

        filter = SequenceFilter(
            name=name,
            metadata=metadata,
            external_id_prefix=external_id_prefix,
            asset_ids=asset_ids,
            asset_subtree_ids=asset_subtree_ids_processed,
            created_time=created_time,
            last_updated_time=last_updated_time,
            data_set_ids=data_set_ids_processed,
        ).dump(camel_case=True)

        prep_sort = prepare_filter_sort(sort, SequenceSort)
        self._validate_filter(advanced_filter)

        async for item in self._list_generator(
            list_cls=SequenceList,
            resource_cls=Sequence,
            method="POST",
            chunk_size=chunk_size,
            filter=filter,
            advanced_filter=advanced_filter,
            limit=limit,
            sort=prep_sort,
        ):
            yield item

    async def retrieve(self, id: int | None = None, external_id: str | None = None) -> Sequence | None:
        """`Retrieve a single sequence by id. <https://developer.cognite.com/api#tag/Sequences/operation/getSequenceById>`_

        Args:
            id (int | None): ID
            external_id (str | None): External ID

        Returns:
            Sequence | None: Requested sequence or None if it does not exist.

        Examples:

            Get sequence by id:

                >>> from cognite.client import CogniteClient, AsyncCogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> res = client.sequences.retrieve(id=1)

            Get sequence by external id:

                >>> res = client.sequences.retrieve()
        """
        identifiers = IdentifierSequence.load(ids=id, external_ids=external_id).as_singleton()
        return await self._retrieve_multiple(list_cls=SequenceList, resource_cls=Sequence, identifiers=identifiers)

    async def retrieve_multiple(
        self,
        ids: typing.Sequence[int] | None = None,
        external_ids: SequenceNotStr[str] | None = None,
        ignore_unknown_ids: bool = False,
    ) -> SequenceList:
        """`Retrieve multiple sequences by id. <https://developer.cognite.com/api#tag/Sequences/operation/getSequenceById>`_

        Args:
            ids (typing.Sequence[int] | None): IDs
            external_ids (SequenceNotStr[str] | None): External IDs
            ignore_unknown_ids (bool): Ignore IDs and external IDs that are not found rather than throw an exception.

        Returns:
            SequenceList: The requested sequences.

        Examples:

            Get sequences by id:

                >>> from cognite.client import CogniteClient, AsyncCogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> res = client.sequences.retrieve_multiple(ids=[1, 2, 3])

            Get sequences by external id:

                >>> res = client.sequences.retrieve_multiple(external_ids=["abc", "def"])
        """
        identifiers = IdentifierSequence.load(ids=ids, external_ids=external_ids)
        return await self._retrieve_multiple(
            list_cls=SequenceList, resource_cls=Sequence, identifiers=identifiers, ignore_unknown_ids=ignore_unknown_ids
        )

    async def aggregate_count(
        self,
        advanced_filter: Filter | dict[str, Any] | None = None,
        filter: SequenceFilter | dict[str, Any] | None = None,
    ) -> int:
        """`Count of sequences matching the specified filters and search. <https://developer.cognite.com/api#tag/Sequences/operation/aggregateSequences>`_

        Args:
            advanced_filter (Filter | dict[str, Any] | None): The filter to narrow down the sequences to count.
            filter (SequenceFilter | dict[str, Any] | None): The filter to narrow down sequences to count requiring exact match.

        Returns:
            int: The number of sequences matching the specified filters and search.

        Examples:

            Count the number of time series in your CDF project:

                >>> from cognite.client import CogniteClient, AsyncCogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> count = client.sequences.aggregate_count()

            Count the number of sequences with external id prefixed with "mapping:" in your CDF project:

                >>> from cognite.client.data_classes import filters
                >>> from cognite.client.data_classes.sequences import SequenceProperty
                >>> is_mapping = filters.Prefix(SequenceProperty.external_id, "mapping:")
                >>> count = client.sequences.aggregate_count(advanced_filter=is_mapping)

        """
        self._validate_filter(advanced_filter)
        return await self._advanced_aggregate(
            "count",
            filter=filter,
            advanced_filter=advanced_filter,
            api_subversion="beta",
        )

    async def aggregate_cardinality_values(
        self,
        property: SequenceProperty | str | list[str],
        advanced_filter: Filter | dict[str, Any] | None = None,
        aggregate_filter: AggregationFilter | dict[str, Any] | None = None,
        filter: SequenceFilter | dict[str, Any] | None = None,
    ) -> int:
        """`Find approximate property count for sequences. <https://developer.cognite.com/api#tag/Sequences/operation/aggregateSequences>`_

        Args:
            property (SequenceProperty | str | list[str]): The property to count the cardinality of.
            advanced_filter (Filter | dict[str, Any] | None): The filter to narrow down the sequences to count cardinality.
            aggregate_filter (AggregationFilter | dict[str, Any] | None): The filter to apply to the resulting buckets.
            filter (SequenceFilter | dict[str, Any] | None): The filter to narrow down the sequences  to count requiring exact match.

        Returns:
            int: The number of properties matching the specified filters and search.

        Examples:

            Count the number of different values for the metadata key "efficiency" used for sequences in your CDF project:

                >>> from cognite.client import CogniteClient
                >>> from cognite.client.data_classes.sequences import SequenceProperty
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> count = client.sequences.aggregate_cardinality_values(SequenceProperty.metadata_key("efficiency"))

            Count the number of timezones (metadata key) for sequences with the word "critical" in the description
            in your CDF project, but exclude timezones from america:

                >>> from cognite.client.data_classes import filters, aggregations as aggs
                >>> from cognite.client.data_classes.sequences import SequenceProperty
                >>> not_america = aggs.Not(aggs.Prefix("america"))
                >>> is_critical = filters.Search(SequenceProperty.description, "critical")
                >>> timezone_count = client.sequences.aggregate_cardinality_values(
                ...     SequenceProperty.metadata_key("timezone"),
                ...     advanced_filter=is_critical,
                ...     aggregate_filter=not_america)
        """
        self._validate_filter(advanced_filter)
        return await self._advanced_aggregate(
            "cardinalityValues",
            properties=property,
            filter=filter,
            advanced_filter=advanced_filter,
            aggregate_filter=aggregate_filter,
            api_subversion="beta",
        )

    async def aggregate_cardinality_properties(
        self,
        path: SequenceProperty | str | list[str],
        advanced_filter: Filter | dict[str, Any] | None = None,
        aggregate_filter: AggregationFilter | dict[str, Any] | None = None,
        filter: SequenceFilter | dict[str, Any] | None = None,
    ) -> int:
        """`Find approximate paths count for sequences.  <https://developer.cognite.com/api#tag/Sequences/operation/aggregateSequences>`_

        Args:
            path (SequenceProperty | str | list[str]): The scope in every document to aggregate properties. The only value allowed now is ["metadata"]. It means to aggregate only metadata properties (aka keys).
            advanced_filter (Filter | dict[str, Any] | None): The filter to narrow down the sequences to count cardinality.
            aggregate_filter (AggregationFilter | dict[str, Any] | None): The filter to apply to the resulting buckets.
            filter (SequenceFilter | dict[str, Any] | None): The filter to narrow down the sequences  to count requiring exact match.

        Returns:
            int: The number of properties matching the specified filters and search.

        Examples:

            Count the number of different metadata keys in your CDF project:

                >>> from cognite.client import CogniteClient
                >>> from cognite.client.data_classes.sequences import SequenceProperty
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> count = client.sequences.aggregate_cardinality_values(SequenceProperty.metadata)
        """
        self._validate_filter(advanced_filter)
        return await self._advanced_aggregate(
            "cardinalityProperties",
            path=path,
            filter=filter,
            advanced_filter=advanced_filter,
            aggregate_filter=aggregate_filter,
            api_subversion="beta",
        )

    async def aggregate_unique_values(
        self,
        property: SequenceProperty | str | list[str],
        advanced_filter: Filter | dict[str, Any] | None = None,
        aggregate_filter: AggregationFilter | dict[str, Any] | None = None,
        filter: SequenceFilter | dict[str, Any] | None = None,
    ) -> UniqueResultList:
        """`Get unique paths with counts for sequences. <https://developer.cognite.com/api#tag/Sequences/operation/aggregateSequences>`_

        Args:
            property (SequenceProperty | str | list[str]): The property to group by.
            advanced_filter (Filter | dict[str, Any] | None): The filter to narrow down the sequences to count cardinality.
            aggregate_filter (AggregationFilter | dict[str, Any] | None): The filter to apply to the resulting buckets.
            filter (SequenceFilter | dict[str, Any] | None): The filter to narrow down the sequences to count requiring exact match.

        Returns:
            UniqueResultList: List of unique values of sequences matching the specified filters and search.

        Examples:

            Get the timezones (metadata key) with count for your sequences in your CDF project:

                >>> from cognite.client import CogniteClient
                >>> from cognite.client.data_classes.sequences import SequenceProperty
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> result = client.sequences.aggregate_unique_values(SequenceProperty.metadata_key("timezone"))
                >>> print(result.unique)

            Get the different metadata keys with count used for sequences created after 2020-01-01 in your CDF project:

                >>> from cognite.client.data_classes import filters
                >>> from cognite.client.data_classes.sequences import SequenceProperty
                >>> from cognite.client.utils import timestamp_to_ms
                >>> from datetime import datetime
                >>> created_after_2020 = filters.Range(SequenceProperty.created_time, gte=timestamp_to_ms(datetime(2020, 1, 1)))
                >>> result = client.sequences.aggregate_unique_values(SequenceProperty.metadata, advanced_filter=created_after_2020)
                >>> print(result.unique)

            Get the different metadata keys with count for sequences updated after 2020-01-01 in your CDF project, but exclude all metadata keys that
            starts with "test":

                >>> from cognite.client.data_classes.sequences import SequenceProperty
                >>> from cognite.client.data_classes import aggregations as aggs, filters
                >>> not_test = aggs.Not(aggs.Prefix("test"))
                >>> created_after_2020 = filters.Range(SequenceProperty.last_updated_time, gte=timestamp_to_ms(datetime(2020, 1, 1)))
                >>> result = client.sequences.aggregate_unique_values(SequenceProperty.metadata, advanced_filter=created_after_2020, aggregate_filter=not_test)
                >>> print(result.unique)
        """
        self._validate_filter(advanced_filter)
        if property == ["metadata"] or property is SequenceProperty.metadata:
            return await self._advanced_aggregate(
                aggregate="uniqueProperties",
                path=property,
                filter=filter,
                advanced_filter=advanced_filter,
                aggregate_filter=aggregate_filter,
                api_subversion="beta",
            )
        return await self._advanced_aggregate(
            aggregate="uniqueValues",
            properties=property,
            filter=filter,
            advanced_filter=advanced_filter,
            aggregate_filter=aggregate_filter,
            api_subversion="beta",
        )

    async def aggregate_unique_properties(
        self,
        path: SequenceProperty | str | list[str],
        advanced_filter: Filter | dict[str, Any] | None = None,
        aggregate_filter: AggregationFilter | dict[str, Any] | None = None,
        filter: SequenceFilter | dict[str, Any] | None = None,
    ) -> UniqueResultList:
        """`Find approximate unique sequence properties. <https://developer.cognite.com/api#tag/Sequences/operation/aggregateSequences>`_

        Args:
            path (SequenceProperty | str | list[str]): The scope in every document to aggregate properties. The only value allowed now is ["metadata"]. It means to aggregate only metadata properties (aka keys).
            advanced_filter (Filter | dict[str, Any] | None): The filter to narrow down the sequences to count cardinality.
            aggregate_filter (AggregationFilter | dict[str, Any] | None): The filter to apply to the resulting buckets.
            filter (SequenceFilter | dict[str, Any] | None): The filter to narrow down the sequences to count requiring exact match.

        Returns:
            UniqueResultList: List of unique values of sequences matching the specified filters and search.

        Examples:

            Get the metadata keys with count for your sequences in your CDF project:

                >>> from cognite.client import CogniteClient
                >>> from cognite.client.data_classes.sequences import SequenceProperty
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> result = client.sequences.aggregate_unique_properties(SequenceProperty.metadata)
        """
        self._validate_filter(advanced_filter)
        return await self._advanced_aggregate(
            aggregate="uniqueProperties",
            path=path,
            filter=filter,
            advanced_filter=advanced_filter,
            aggregate_filter=aggregate_filter,
            api_subversion="beta",
        )

    @overload
    async def create(self, sequence: Sequence | SequenceWrite) -> Sequence: ...

    @overload
    async def create(self, sequence: typing.Sequence[Sequence] | typing.Sequence[SequenceWrite]) -> SequenceList: ...

    async def create(
        self, sequence: Sequence | SequenceWrite | typing.Sequence[Sequence] | typing.Sequence[SequenceWrite]
    ) -> Sequence | SequenceList:
        """`Create one or more sequences. <https://developer.cognite.com/api#tag/Sequences/operation/createSequence>`_

        Args:
            sequence (Sequence | SequenceWrite | typing.Sequence[Sequence] | typing.Sequence[SequenceWrite]): Sequence or list of Sequence to create. The Sequence columns parameter is a list of objects with fields `externalId` (external id of the column, when omitted, they will be given ids of 'column0, column1, ...'), `valueType` (data type of the column, either STRING, LONG, or DOUBLE, with default DOUBLE), `name`, `description`, `metadata` (optional fields to describe and store information about the data in the column). Other fields will be removed automatically, so a columns definition from a different sequence object can be passed here.

        Returns:
            Sequence | SequenceList: The created sequence(s).

        Examples:

            Create a new sequence:

                >>> from cognite.client import CogniteClient
                >>> from cognite.client.data_classes import SequenceWrite, SequenceColumnWrite
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> column_def = [
                ...     SequenceColumnWrite(value_type="STRING", external_id="user", description="some description"),
                ...     SequenceColumnWrite(value_type="DOUBLE", external_id="amount")
                ... ]
                >>> seq = client.sequences.create(SequenceWrite(external_id="my_sequence", columns=column_def))

            Create a new sequence with the same column specifications as an existing sequence:

                >>> seq2 = client.sequences.create(SequenceWrite(external_id="my_copied_sequence", columns=column_def))

        """
        assert_type(sequence, "sequences", [typing.Sequence, SequenceCore])

        return await self._create_multiple(
            list_cls=SequenceList, resource_cls=Sequence, items=sequence, input_resource_cls=SequenceWrite
        )

    async def delete(
        self,
        id: int | typing.Sequence[int] | None = None,
        external_id: str | SequenceNotStr[str] | None = None,
        ignore_unknown_ids: bool = False,
    ) -> None:
        """`Delete one or more sequences. <https://developer.cognite.com/api#tag/Sequences/operation/deleteSequences>`_

        Args:
            id (int | typing.Sequence[int] | None): Id or list of ids
            external_id (str | SequenceNotStr[str] | None): External ID or list of external ids
            ignore_unknown_ids (bool): Ignore IDs and external IDs that are not found rather than throw an exception.

        Examples:

            Delete sequences by id or external id:

                >>> from cognite.client import CogniteClient, AsyncCogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> client.sequences.delete(id=[1,2,3], external_id="3")
        """
        await self._delete_multiple(
            identifiers=IdentifierSequence.load(ids=id, external_ids=external_id),
            wrap_ids=True,
            extra_body_fields={"ignoreUnknownIds": ignore_unknown_ids},
        )

    @overload
    async def update(
        self,
        item: Sequence | SequenceWrite | SequenceUpdate,
        mode: Literal["replace_ignore_null", "patch", "replace"] = "replace_ignore_null",
    ) -> Sequence: ...

    @overload
    async def update(
        self,
        item: typing.Sequence[Sequence | SequenceWrite | SequenceUpdate],
        mode: Literal["replace_ignore_null", "patch", "replace"] = "replace_ignore_null",
    ) -> SequenceList: ...

    async def update(
        self,
        item: Sequence | SequenceWrite | SequenceUpdate | typing.Sequence[Sequence | SequenceWrite | SequenceUpdate],
        mode: Literal["replace_ignore_null", "patch", "replace"] = "replace_ignore_null",
    ) -> Sequence | SequenceList:
        """`Update one or more sequences. <https://developer.cognite.com/api#tag/Sequences/operation/updateSequences>`_

        Args:
            item (Sequence | SequenceWrite | SequenceUpdate | typing.Sequence[Sequence | SequenceWrite | SequenceUpdate]): Sequences to update
            mode (Literal['replace_ignore_null', 'patch', 'replace']): How to update data when a non-update object is given (Sequence or -Write). If you use 'replace_ignore_null', only the fields you have set will be used to replace existing (default). Using 'replace' will additionally clear all the fields that are not specified by you. Last option, 'patch', will update only the fields you have set and for container-like fields such as metadata or labels, add the values to the existing. For more details, see :ref:`appendix-update`.

        Returns:
            Sequence | SequenceList: Updated sequences.

        Examples:

            Update a sequence that you have fetched. This will perform a full update of the sequences:

                >>> from cognite.client import CogniteClient, AsyncCogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> res = client.sequences.retrieve(id=1)
                >>> res.description = "New description"
                >>> res = client.sequences.update(res)

            Perform a partial update on a sequence, updating the description and adding a new field to metadata:

                >>> from cognite.client.data_classes import SequenceUpdate
                >>> my_update = SequenceUpdate(id=1).description.set("New description").metadata.add({"key": "value"})
                >>> res = client.sequences.update(my_update)

            **Updating column definitions**

            Currently, updating the column definitions of a sequence is only supported through partial update, using `add`, `remove` and `modify` methods on the `columns` property.

            Add a single new column:

                >>> from cognite.client.data_classes import SequenceUpdate, SequenceColumnWrite
                >>>
                >>> my_update = SequenceUpdate(id=1).columns.add(
                ...     SequenceColumnWrite(value_type ="STRING",external_id="user", description ="some description")
                ... )
                >>> res = client.sequences.update(my_update)

            Add multiple new columns:

                >>> from cognite.client.data_classes import SequenceUpdate, SequenceColumnWrite
                >>>
                >>> column_def = [
                ...     SequenceColumnWrite(value_type ="STRING",external_id="user", description ="some description"),
                ...     SequenceColumnWrite(value_type="DOUBLE", external_id="amount")
                ... ]
                >>> my_update = SequenceUpdate(id=1).columns.add(column_def)
                >>> res = client.sequences.update(my_update)

            Remove a single column:

                >>> from cognite.client.data_classes import SequenceUpdate
                >>>
                >>> my_update = SequenceUpdate(id=1).columns.remove("col_external_id1")
                >>> res = client.sequences.update(my_update)

            Remove multiple columns:

                >>> from cognite.client.data_classes import SequenceUpdate
                >>>
                >>> my_update = SequenceUpdate(id=1).columns.remove(["col_external_id1","col_external_id2"])
                >>> res = client.sequences.update(my_update)

            Update existing columns:

                >>> from cognite.client.data_classes import SequenceUpdate, SequenceColumnUpdate
                >>>
                >>> column_updates = [
                ...     SequenceColumnUpdate(external_id="col_external_id_1").external_id.set("new_col_external_id"),
                ...     SequenceColumnUpdate(external_id="col_external_id_2").description.set("my new description"),
                ... ]
                >>> my_update = SequenceUpdate(id=1).columns.modify(column_updates)
                >>> res = client.sequences.update(my_update)
        """
        cdf_item_by_id = await self._get_cdf_item_by_id(item, "updating")
        return await self._update_multiple(
            list_cls=SequenceList,
            resource_cls=Sequence,
            update_cls=SequenceUpdate,
            items=item,
            mode=mode,
            cdf_item_by_id=cdf_item_by_id,
        )

    @overload
    async def upsert(
        self, item: typing.Sequence[Sequence | SequenceWrite], mode: Literal["patch", "replace"] = "patch"
    ) -> SequenceList: ...

    @overload
    async def upsert(self, item: Sequence | SequenceWrite, mode: Literal["patch", "replace"] = "patch") -> Sequence: ...

    async def upsert(
        self,
        item: Sequence | SequenceWrite | typing.Sequence[Sequence | SequenceWrite],
        mode: Literal["patch", "replace"] = "patch",
    ) -> Sequence | SequenceList:
        """Upsert sequences, i.e., update if it exists, and create if it does not exist.
            Note this is a convenience method that handles the upserting for you by first calling update on all items,
            and if any of them fail because they do not exist, it will create them instead.

            For more details, see :ref:`appendix-upsert`.

        Args:
            item (Sequence | SequenceWrite | typing.Sequence[Sequence | SequenceWrite]): Sequence or list of sequences to upsert.
            mode (Literal['patch', 'replace']): Whether to patch or replace in the case the sequences are existing. If you set 'patch', the call will only update fields with non-null values (default). Setting 'replace' will unset any fields that are not specified.

        Returns:
            Sequence | SequenceList: The upserted sequence(s).

        Examples:

            Upsert for sequences:

                >>> from cognite.client import CogniteClient
                >>> from cognite.client.data_classes import SequenceWrite, SequenceColumnWrite
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> existing_sequence = client.sequences.retrieve(id=1)
                >>> existing_sequence.description = "New description"
                >>> new_sequence = SequenceWrite(
                ...     external_id="new_sequence",
                ...     description="New sequence",
                ...     columns=[SequenceColumnWrite(external_id="col1", value_type="STRING")]
                ... )
                >>> res = client.sequences.upsert([existing_sequence, new_sequence], mode="replace")
        """

        cdf_item_by_id = await self._get_cdf_item_by_id(item, "upserting")
        return await self._upsert_multiple(
            item,
            list_cls=SequenceList,
            resource_cls=Sequence,
            update_cls=SequenceUpdate,
            input_resource_cls=Sequence,
            mode=mode,
            cdf_item_by_id=cdf_item_by_id,
        )

    async def _get_cdf_item_by_id(
        self,
        item: Sequence | SequenceWrite | SequenceUpdate | typing.Sequence[Sequence | SequenceWrite | SequenceUpdate],
        operation: Literal["updating", "upserting"],
    ) -> Mapping[str | int, Sequence]:
        if isinstance(item, SequenceWrite):
            if item.external_id is None:
                raise ValueError(f"External ID must be set when {operation} a SequenceWrite object.")
            cdf_item = await self.retrieve(external_id=item.external_id)
            if cdf_item and cdf_item.external_id:
                return {cdf_item.external_id: cdf_item}
        elif isinstance(item, Sequence):
            if item.external_id:
                cdf_item = await self.retrieve(external_id=item.external_id)
                if cdf_item and cdf_item.external_id:
                    return {cdf_item.external_id: cdf_item}
            else:
                cdf_item = await self.retrieve(id=item.id)
                if cdf_item and cdf_item.id:
                    return {cdf_item.id: cdf_item}
        elif isinstance(item, collections.abc.Sequence):
            external_ids = [i.external_id for i in item if isinstance(i, SequenceWrite)]
            if None in external_ids:
                raise ValueError(f"External ID must be set when {operation} using one or more SequenceWrite object(s).")
            external_ids.extend(i.external_id for i in item if isinstance(i, Sequence) and i.external_id)
            internal_ids = [i.id for i in item if isinstance(i, Sequence) and not i.external_id]
            cdf_items = await self.retrieve_multiple(
                ids=internal_ids, external_ids=typing.cast(list[str], external_ids), ignore_unknown_ids=True
            )
            return {item.external_id if item.external_id is not None else item.id: item for item in cdf_items}
        return {}

    @classmethod
    def _convert_resource_to_patch_object(
        cls,
        resource: CogniteResource,
        update_attributes: list[PropertySpec],
        mode: Literal["replace_ignore_null", "patch", "replace"] = "replace_ignore_null",
        cdf_item_by_id: Mapping[Any, Sequence] | None = None,  # type: ignore [override]
    ) -> dict[str, dict[str, dict]]:
        update_obj = super()._convert_resource_to_patch_object(resource, update_attributes, mode)
        if not isinstance(resource, SequenceWrite | Sequence):
            return update_obj
        # Lookup columns to check what to add, remove and modify
        cdf_item: Sequence | None = None
        if cdf_item_by_id:
            if isinstance(resource, Sequence | SequenceWrite) and resource.external_id in cdf_item_by_id:
                cdf_item = cdf_item_by_id[cast(str, resource.external_id)]
            elif isinstance(resource, Sequence) and resource.id in cdf_item_by_id:
                cdf_item = cdf_item_by_id[resource.id]
        if isinstance(resource, Sequence):
            if resource.columns is None:
                resource_columns = SequenceColumnWriteList([])
            else:
                resource_columns = resource.columns.as_write()
        else:
            resource_columns = resource.columns

        update_obj["update"]["columns"] = {}
        if cdf_item is None:
            update_obj["update"]["columns"]["add"] = [column.dump() for column in resource_columns]
            return update_obj

        cdf_columns_by_external_id = {
            column.external_id: column for column in cdf_item.columns or [] if column.external_id
        }
        columns_by_external_id = {column.external_id: column for column in resource_columns if column.external_id}
        to_remove = set(cdf_columns_by_external_id.keys()) - set(columns_by_external_id.keys())
        if mode != "patch" and to_remove:
            # Replace or replace_ignore_null, remove all columns that are not in the new columns
            update_obj["update"]["columns"]["remove"] = list(to_remove)
        if to_add := (set(columns_by_external_id.keys()) - set(cdf_columns_by_external_id.keys())):
            update_obj["update"]["columns"]["add"] = [columns_by_external_id[ext_id].dump() for ext_id in to_add]
        if to_modify := (set(columns_by_external_id.keys()) & set(cdf_columns_by_external_id.keys())):
            modify_list: list[dict[str, dict[str, Any]]] = []
            for col_ext_id in to_modify:
                col_write_obj = columns_by_external_id[col_ext_id]
                column_update = super()._convert_resource_to_patch_object(
                    col_write_obj,
                    SequenceColumnUpdate._get_update_properties(col_write_obj),
                    mode,
                )
                modify_list.append(column_update)
            update_obj["update"]["columns"]["modify"] = modify_list
        return update_obj

    async def search(
        self,
        name: str | None = None,
        description: str | None = None,
        query: str | None = None,
        filter: SequenceFilter | dict[str, Any] | None = None,
        limit: int = DEFAULT_LIMIT_READ,
    ) -> SequenceList:
        """`Search for sequences. <https://developer.cognite.com/api#tag/Sequences/operation/searchSequences>`_
        Primarily meant for human-centric use-cases and data exploration, not for programs, since matching and ordering may change over time. Use the `list` function if stable or exact matches are required.

        Args:
            name (str | None): Prefix and fuzzy search on name.
            description (str | None): Prefix and fuzzy search on description.
            query (str | None): Search on name and description using wildcard search on each of the words (separated by spaces). Retrieves results where at least one word must match. Example: 'some other'
            filter (SequenceFilter | dict[str, Any] | None): Filter to apply. Performs exact match on these fields.
            limit (int): Max number of results to return.

        Returns:
            SequenceList: The search result as a SequenceList

        Examples:

            Search for a sequence:

                >>> from cognite.client import CogniteClient, AsyncCogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> res = client.sequences.search(name="some name")
        """
        return await self._search(
            list_cls=SequenceList,
            search={"name": name, "description": description, "query": query},
            filter=filter or {},
            limit=limit,
        )

    def _validate_filter(self, filter: Filter | dict[str, Any] | None) -> None:
        _validate_filter(filter, _FILTERS_SUPPORTED, type(self).__name__)

    async def list(
        self,
        name: str | None = None,
        external_id_prefix: str | None = None,
        metadata: dict[str, str] | None = None,
        asset_ids: typing.Sequence[int] | None = None,
        asset_subtree_ids: int | typing.Sequence[int] | None = None,
        asset_subtree_external_ids: str | SequenceNotStr[str] | None = None,
        data_set_ids: int | typing.Sequence[int] | None = None,
        data_set_external_ids: str | SequenceNotStr[str] | None = None,
        created_time: dict[str, Any] | TimestampRange | None = None,
        last_updated_time: dict[str, Any] | TimestampRange | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        partitions: int | None = None,
        advanced_filter: Filter | dict[str, Any] | None = None,
        sort: SortSpec | list[SortSpec] | None = None,
    ) -> SequenceList:
        """`List sequences <https://developer.cognite.com/api#tag/Sequences/operation/advancedListSequences>`_

        Args:
            name (str | None): Filter out sequences that do not have this *exact* name.
            external_id_prefix (str | None): Filter out sequences that do not have this string as the start of the externalId
            metadata (dict[str, str] | None): Filter out sequences that do not match these metadata fields and values (case-sensitive). Format is {"key1":"value1","key2":"value2"}.
            asset_ids (typing.Sequence[int] | None): Filter out sequences that are not linked to any of these assets.
            asset_subtree_ids (int | typing.Sequence[int] | None): Only include sequences that have a related asset in a subtree rooted at any of these assetIds. If the total size of the given subtrees exceeds 100,000 assets, an error will be returned.
            asset_subtree_external_ids (str | SequenceNotStr[str] | None): Only include sequences that have a related asset in a subtree rooted at any of these assetExternalIds. If the total size of the given subtrees exceeds 100,000 assets, an error will be returned.
            data_set_ids (int | typing.Sequence[int] | None): Return only sequences in the specified data set(s) with this id / these ids.
            data_set_external_ids (str | SequenceNotStr[str] | None): Return only sequences in the specified data set(s) with this external id / these external ids.
            created_time (dict[str, Any] | TimestampRange | None):  Range between two timestamps. Possible keys are `min` and `max`, with values given as time stamps in ms.
            last_updated_time (dict[str, Any] | TimestampRange | None):  Range between two timestamps. Possible keys are `min` and `max`, with values given as time stamps in ms.
            limit (int | None): Max number of sequences to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            partitions (int | None): Retrieve resources in parallel using this number of workers (values up to 10 allowed), limit must be set to `None` (or `-1`).
            advanced_filter (Filter | dict[str, Any] | None): Advanced filter query using the filter DSL (Domain Specific Language). It allows defining complex filtering expressions that combine simple operations, such as equals, prefix, exists, etc., using boolean operators and, or, and not. See examples below for usage.
            sort (SortSpec | list[SortSpec] | None): The criteria to sort by. Defaults to desc for `_score_` and asc for all other properties. Sort is not allowed if `partitions` is used.

        Returns:
            SequenceList: The requested sequences.

        .. note::
            When using `partitions`, there are few considerations to keep in mind:
                * `limit` has to be set to `None` (or `-1`).
                * API may reject requests if you specify more than 10 partitions. When Cognite enforces this behavior, the requests result in a 400 Bad Request status.
                * Partitions are done independently of sorting: there's no guarantee of the sort order between elements from different partitions. For this reason providing a `sort` parameter when using `partitions` is not allowed.


        Examples:

            List sequences:

                >>> from cognite.client import CogniteClient, AsyncCogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> res = client.sequences.list(limit=5)

            Iterate over sequences, one-by-one:

                >>> for seq in client.sequences():
                ...     seq  # do something with the sequence

            Iterate over chunks of sequences to reduce memory load:

                >>> for seq_list in client.sequences(chunk_size=2500):
                ...     seq_list # do something with the sequences

            Using advanced filter, find all sequences that have a metadata key 'timezone' starting with 'Europe',
            and sort by external id ascending:

                >>> from cognite.client.data_classes import filters
                >>> in_timezone = filters.Prefix(["metadata", "timezone"], "Europe")
                >>> res = client.sequences.list(advanced_filter=in_timezone, sort=("external_id", "asc"))

            Note that you can check the API documentation above to see which properties you can filter on
            with which filters.

            To make it easier to avoid spelling mistakes and easier to look up available properties
            for filtering and sorting, you can also use the `SequenceProperty` and `SortableSequenceProperty` Enums.

                >>> from cognite.client.data_classes import filters
                >>> from cognite.client.data_classes.sequences import SequenceProperty, SortableSequenceProperty
                >>> in_timezone = filters.Prefix(SequenceProperty.metadata_key("timezone"), "Europe")
                >>> res = client.sequences.list(
                ...     advanced_filter=in_timezone,
                ...     sort=(SortableSequenceProperty.external_id, "asc"))

            Combine filter and advanced filter:

                >>> from cognite.client.data_classes import filters
                >>> not_instrument_lvl5 = filters.And(
                ...    filters.ContainsAny("labels", ["Level5"]),
                ...    filters.Not(filters.ContainsAny("labels", ["Instrument"]))
                ... )
                >>> res = client.sequences.list(asset_subtree_ids=[123456], advanced_filter=not_instrument_lvl5)

        """
        asset_subtree_ids_processed = process_asset_subtree_ids(asset_subtree_ids, asset_subtree_external_ids)
        data_set_ids_processed = process_data_set_ids(data_set_ids, data_set_external_ids)

        filter = SequenceFilter(
            name=name,
            metadata=metadata,
            external_id_prefix=external_id_prefix,
            asset_ids=asset_ids,
            asset_subtree_ids=asset_subtree_ids_processed,
            created_time=created_time,
            last_updated_time=last_updated_time,
            data_set_ids=data_set_ids_processed,
        ).dump(camel_case=True)

        prep_sort = prepare_filter_sort(sort, SequenceSort)
        self._validate_filter(advanced_filter)

        return await self._list(
            list_cls=SequenceList,
            resource_cls=Sequence,
            method="POST",
            filter=filter,
            advanced_filter=advanced_filter,
            sort=prep_sort,
            limit=limit,
            partitions=partitions,
        )



from __future__ import annotations

from abc import ABC
from functools import cached_property
from urllib.parse import urljoin

from cognite.client._api_client import APIClient
from cognite.client.exceptions import CogniteOrganizationError
from cognite.client.utils._async_helpers import run_sync


class OrgAPIClient(APIClient, ABC):
    _AUTH_URL = "https://auth.cognite.com"

    # TODO: Code smell: _base_url_with_base_path calling _organization calling run_sync in an async-first SDK...
    @property
    def _base_url_with_base_path(self) -> str:
        # The OrganizationAPI uses the auth_url as the base for these endpoints instead of the
        # base_url like the rest of the SDK.
        assert self._api_version, "API version must be set for organization endpoints"
        return urljoin(self._AUTH_URL, f"/api/{self._api_version}/orgs/{self._organization}")

    @cached_property
    def _organization(self) -> str:
        # This is an internal endpoint, not part of the public API - and we call it manually to avoid
        # the overridden _base_url_with_base_path property:
        full_url = urljoin(self._config.base_url, f"/api/{self._api_version}/projects/{self._config.project}")
        try:
            response = run_sync(self._request("GET", full_url=full_url, include_cdf_headers=True))
            return response.json()["organization"]
        except Exception as err:
            raise CogniteOrganizationError from err



from __future__ import annotations

import atexit
import inspect
import json
import tempfile
import threading
import time
import warnings
from abc import abstractmethod
from collections.abc import Callable
from datetime import datetime
from pathlib import Path
from types import MappingProxyType
from typing import Any, Protocol, runtime_checkable

import requests
from authlib.integrations.httpx_client import OAuth2Client, OAuthError
from msal import ConfidentialClientApplication, PublicClientApplication, SerializableTokenCache

from cognite.client.exceptions import CogniteAuthError, CogniteOAuthError
from cognite.client.utils._auxiliary import at_least_one_is_not_none, exactly_one_is_not_none, load_resource_to_dict

_TOKEN_EXPIRY_LEEWAY_SECONDS_DEFAULT = 30  # Do not change without also updating all the docstrings using it


@runtime_checkable
class CredentialProvider(Protocol):
    @abstractmethod
    def authorization_header(self) -> tuple[str, str]:
        raise NotImplementedError

    @classmethod
    def load(cls, config: dict[str, Any] | str) -> CredentialProvider:
        """Load a credential provider object from a YAML/JSON string or dict.

        Note:
            The dictionary must contain exactly one top level key, which is the type of the credential provider and must be one of the
            following strings: ``"token"``, ``"client_credentials"``, ``"interactive"``, ``"device_code"``, ``"client_certificate"``.
            The value of the key is a dictionary containing the configuration for the credential provider.

        Args:
            config (dict[str, Any] | str): A dictionary or YAML/JSON string containing the configuration for the credential provider.

        Returns:
            CredentialProvider: Initialized credential provider of the specified type.

        Examples:

            Get a token credential provider:

                >>> from cognite.client.credentials import CredentialProvider
                >>> config = {"token": "my secret token"}
                >>> credentials = CredentialProvider.load(config)

            Get a client credential provider:

                >>> import os
                >>> config = {
                ...     "client_credentials": {
                ...         "client_id": "abcd",
                ...         "client_secret": os.environ["OAUTH_CLIENT_SECRET"],
                ...         "token_url": "https://login.microsoftonline.com/xyz/oauth2/v2.0/token",
                ...         "scopes": ["https://api.cognitedata.com/.default"],
                ...     }
                ... }
                >>> credentials = CredentialProvider.load(config)
        """
        loaded = load_resource_to_dict(config)

        if len(loaded) != 1:
            raise ValueError(
                f"Credential provider configuration must be a dictionary containing exactly one of the following "
                f"supported types as the top level key: {sorted(_SUPPORTED_CREDENTIAL_TYPES.keys())}."
            )

        credential_type, credential_config = next(iter(loaded.items()))

        if credential_type not in _SUPPORTED_CREDENTIAL_TYPES:
            raise ValueError(
                f"Invalid credential provider type given, the valid options are: {sorted(_SUPPORTED_CREDENTIAL_TYPES.keys())}."
            )

        if credential_type == "token" and (isinstance(credential_config, str) or callable(credential_config)):
            credential_config = {"token": credential_config}
        return _SUPPORTED_CREDENTIAL_TYPES[credential_type].load(credential_config)  # type: ignore [attr-defined]


class Token(CredentialProvider):
    """Token credential provider

    Args:
        token (str | Callable[[], str]): A token or a token factory.

    Examples:

            >>> from cognite.client.credentials import Token
            >>> token_provider = Token("my secret token")
            >>> token_factory_provider = Token(lambda: "my secret token")

    Note:
        If you pass in a callable, we will expect that you supplied a function that may do a token refresh
        under the hood, so it will be called while holding a thread lock (threading.Lock()).
    """

    def __init__(self, token: str | Callable[[], str]) -> None:
        if isinstance(token, str):
            self.__token_factory = lambda: token

        elif callable(token):  # mypy flat out refuses variations of: isinstance(token, collections.abc.Callable)
            token_refresh_lock = threading.Lock()

            def thread_safe_get_token() -> str:
                assert not isinstance(token, str)  # unbelievable
                with token_refresh_lock:
                    return token()

            self.__token_factory = thread_safe_get_token
        else:
            raise TypeError(f"'token' must be a string or a no-argument-callable returning a string, not {type(token)}")

    def authorization_header(self) -> tuple[str, str]:
        return "Authorization", f"Bearer {self.__token_factory()}"

    @classmethod
    def load(cls, config: dict[str, str | Callable[[], str]] | str) -> Token:
        """Load a token credential provider object from a YAML/JSON string or dict.

        Args:
            config (dict[str, str | Callable[[], str]] | str): A dictionary or YAML/JSON string containing configuration values defined in the Token class.

        Returns:
            Token: Initialized token credential provider.

        Note:
            A callable token is not supported if passing in a yaml string.

        Examples:

            >>> from cognite.client.credentials import Token
            >>> credentials = Token.load({"token": "my secret token"})
        """
        loaded = load_resource_to_dict(config)
        return cls(token=loaded["token"])


class _OAuthCredentialProviderWithTokenRefresh(CredentialProvider):
    def __init__(self, token_expiry_leeway_seconds: int = _TOKEN_EXPIRY_LEEWAY_SECONDS_DEFAULT) -> None:
        # This ensures we don't return a token which expires immediately:
        self.token_expiry_leeway_seconds = token_expiry_leeway_seconds

        self.__token_refresh_lock = threading.Lock()
        self.__access_token: str | None = None
        self.__access_token_expires_at: float | None = None

    def __getstate__(self) -> dict[str, Any]:
        # threading.Lock is not picklable, temporarily remove:
        lock_tmp, self.__token_refresh_lock = self.__token_refresh_lock, None  # type: ignore [assignment]
        state = self.__dict__.copy()
        self.__token_refresh_lock = lock_tmp
        return state

    def __setstate__(self, state: dict[str, Any]) -> None:
        self.__dict__ = state
        self.__token_refresh_lock = threading.Lock()

    @abstractmethod
    def _refresh_access_token(self) -> tuple[str, float]:
        """This method should return the access_token and time until expiration (expire_in)"""
        raise NotImplementedError

    def __should_refresh_token(self, token: str | None, expires_at: float | None) -> bool:
        return token is None or expires_at is None or expires_at < time.time() + self.token_expiry_leeway_seconds

    @staticmethod
    def _verify_credentials(credentials: dict[str, Any]) -> None:
        """The msal library doesn't raise anything when auth fails, but returns a dictionary with varying keys"""
        if "access_token" in credentials and "expires_in" in credentials:
            return

        # 'error_description' includes Windows-style newlines \r\n meant to print nicely. Prettify for exception:
        err_descr = " ".join(credentials.get("error_description", "").splitlines())
        raise CogniteOAuthError(credentials["error"], err_descr)

    def authorization_header(self) -> tuple[str, str]:
        # We lock here to ensure we don't issue a herd of refresh requests in concurrent scenarios.
        # This means we will block all requests for the duration of the token refresh.
        # TODO: Consider instead having a background thread periodically refresh the token to avoid this blocking.
        with self.__token_refresh_lock:
            if self.__should_refresh_token(self.__access_token, self.__access_token_expires_at):
                self.__access_token, self.__access_token_expires_at = self._refresh_access_token()
        return "Authorization", f"Bearer {self.__access_token}"


class _WithMsalSerializableTokenCache:
    @staticmethod
    def _create_serializable_token_cache(cache_path: Path, clear_cache: bool = False) -> SerializableTokenCache:
        token_cache = SerializableTokenCache()

        if cache_path.exists():
            if clear_cache:
                cache_path.unlink(missing_ok=True)
            else:
                with cache_path.open() as fh:
                    token_cache.deserialize(fh.read())

        def __at_exit() -> None:
            if token_cache.has_state_changed:
                with open(cache_path, "w+") as fh:
                    fh.write(token_cache.serialize())

        atexit.register(__at_exit)
        return token_cache

    @staticmethod
    def _resolve_token_cache_path(token_cache_path: Path | None, client_id: str) -> Path:
        return token_cache_path or Path(tempfile.gettempdir()) / f"cognitetokencache.{client_id}.bin"

    def _create_client_app(
        self,
        token_cache_path: Path,
        client_id: str,
        authority_url: str | None = None,
        oauth_discovery_url: str | None = None,
        clear_cache: bool = False,
        mem_cache_only: bool = False,
    ) -> PublicClientApplication:
        from cognite.client.config import global_config

        if authority_url and oauth_discovery_url:
            raise ValueError(
                "Only one of 'authority_url' (for MS Entra) or 'oauth_discovery_url' (for other IdPs) should be provided."
            )

        # In addition to caching in memory, we also cache the token on disk so it can be reused across processes.
        if mem_cache_only:
            serializable_token_cache = SerializableTokenCache()
        else:
            serializable_token_cache = self._create_serializable_token_cache(token_cache_path, clear_cache)
        return PublicClientApplication(
            client_id=client_id,
            authority=authority_url,
            token_cache=serializable_token_cache,
            verify=not global_config.disable_ssl,
            oidc_authority=oauth_discovery_url,
            # These two must be set to `False` to support non-Entra authorities.
            instance_discovery=False,
            validate_authority=False,
        )

    @staticmethod
    def _get_cached_token(cache_path: Path) -> dict[str, Any]:
        if not cache_path.exists():
            return {}
        token = json.loads(cache_path.read_text())
        return token


class OAuthDeviceCode(_OAuthCredentialProviderWithTokenRefresh, _WithMsalSerializableTokenCache):
    """OAuth credential provider for the device code login flow.

    Args:
        authority_url (str | None): MS Entra OAuth authority url, typically "https://login.microsoftonline.com/{tenant_id}"
        client_id (str): Your application's client id that allows device code flows.
        scopes (list[str] | None): A list of scopes.
        cdf_cluster (str | None): The CDF cluster where the CDF project is located. If provided, scopes will be set to
            [f"https://{cdf_cluster}.cognitedata.com/IDENTITY https://{cdf_cluster}.cognitedata.com/user_impersonation openid profile"].
        oauth_discovery_url (str | None): Standard OAuth discovery URL, should be where "/.well-known/openid-configuration" is found.
        token_cache_path (Path | None): Location to store token cache, defaults to os temp directory/cognitetokencache.{client_id}.bin.
        token_expiry_leeway_seconds (int): The token is refreshed at the earliest when this number of seconds is left before expiry. Default: 30 sec
        clear_cache (bool): If True, the token cache will be cleared on initialization. Default: False
        mem_cache_only (bool): If True, the token cache will only be stored in memory. Default: False
        **token_custom_args (Any): Additional request parameters to pass to the authorization endpoint.
    Examples:

            >>> from cognite.client.credentials import OAuthDeviceCode
            >>> oauth_provider = OAuthDeviceCode(
            ...     authority_url="https://login.microsoftonline.com/xyz",
            ...     client_id="abcd",
            ...     scopes=["https://greenfield.cognitedata.com/.default"],
            ... )

        Create credentials with auth0

            >>> from cognite.client.credentials import OAuthDeviceCode
            >>> oauth_provider = OAuthDeviceCode(
            ...     authority_url=None,
            ...     oauth_discovery_url="https://my-tenant.auth0.com/oauth",
            ...     client_id="abcd",
            ...     scopes=["IDENTITY", "user_impersonation"],
            ... )
    """

    def __init__(
        self,
        authority_url: str | None,
        client_id: str,
        scopes: list[str] | None = None,
        cdf_cluster: str | None = None,
        oauth_discovery_url: str | None = None,
        token_cache_path: Path | None = None,
        token_expiry_leeway_seconds: int = _TOKEN_EXPIRY_LEEWAY_SECONDS_DEFAULT,
        clear_cache: bool = False,
        mem_cache_only: bool = False,
        **token_custom_args: Any,
    ) -> None:
        super().__init__(token_expiry_leeway_seconds)
        if not exactly_one_is_not_none(authority_url, oauth_discovery_url):
            raise ValueError("Either 'authority_url' or 'oauth_discovery_url' must be provided, and not both.")
        if not at_least_one_is_not_none(scopes, cdf_cluster):
            raise ValueError("Either 'scopes' or 'cdf_cluster' must be provided.")
        if not client_id:
            raise ValueError("'client_id' must be provided.")
        self.__authority_url = authority_url
        self.__oauth_discovery_url = oauth_discovery_url
        self.__client_id = client_id
        self.__scopes = scopes or [
            f"https://{cdf_cluster}.cognitedata.com/IDENTITY",
            f"https://{cdf_cluster}.cognitedata.com/user_impersonation",
            "openid",
            "profile",
        ]
        self.__mem_cache_only = mem_cache_only
        self.__token_custom_args = token_custom_args

        self._token_cache_path = self._resolve_token_cache_path(token_cache_path, client_id)
        self.__app = self._create_client_app(
            self._token_cache_path,
            client_id,
            authority_url,
            oauth_discovery_url,
            clear_cache,
            mem_cache_only,
        )

    def __getstate__(self) -> dict[str, Any]:
        # PublicClientApplication is not picklable, temporarily remove:
        app_tmp, self.__app = self.__app, None
        state = super().__getstate__()
        self.__app = app_tmp
        return state

    def __setstate__(self, state: dict[str, Any]) -> None:
        super().__setstate__(state)
        self.__app = self._create_client_app(
            token_cache_path=self._token_cache_path,
            client_id=self.__client_id,
            authority_url=self.__authority_url,
            oauth_discovery_url=self.__oauth_discovery_url,
            clear_cache=False,
            mem_cache_only=self.__mem_cache_only,
        )

    @property
    def authority_url(self) -> str | None:
        return self.__authority_url

    @property
    def oauth_discovery_url(self) -> str | None:
        return self.__oauth_discovery_url

    @property
    def client_id(self) -> str:
        return self.__client_id

    @property
    def scopes(self) -> list[str]:
        return self.__scopes

    def scope_string(self) -> str:
        return " ".join(self.__scopes)

    def _get_device_authorization_endpoint(self) -> str:
        # First try MSAL authority object
        if self.__app.authority and (
            device_auth_endpoint := getattr(self.__app.authority, "device_authorization_endpoint", None)
        ):
            return device_auth_endpoint

        # Handle Entra ID without MSAL
        if self.__authority_url:
            return self.__authority_url.rstrip("/") + "/oauth2/v2.0/devicecode"

        # Handle standard OIDC discovery (Auth0, Cognito, etc.)
        if not self.__oauth_discovery_url:
            # all options exhausted, raise error
            raise CogniteAuthError("Unable to determine device authorization endpoint")

        # This will attempt to make a request to get OIDC service
        # and parse the response (the discovery document)
        oidc_config_url = self.__oauth_discovery_url.rstrip("/") + "/.well-known/openid-configuration"
        try:
            oidc_response = self.__app.http_client.get(oidc_config_url)
        except requests.exceptions.RequestException as e:
            raise CogniteAuthError(f"Error fetching device_authorization_endpoint from OIDC discovery: {e}") from e

        try:
            oidc_metadata = oidc_response.json()
        except (json.JSONDecodeError, TypeError, AttributeError) as e:
            raise CogniteAuthError(f"Error parsing OIDC discovery document: {e}") from e

        if device_auth_endpoint := oidc_metadata.get("device_authorization_endpoint"):
            return device_auth_endpoint
        raise CogniteAuthError(
            f"device_authorization_endpoint not found in OIDC discovery document at {oidc_config_url}"
        )

    def _get_device_code_response(self, device_auth_endpoint: str, data: dict[str, Any]) -> dict[str, Any]:
        """Initiate device code flow and return the device flow object.

        Args:
            device_auth_endpoint (str): The device authorization endpoint URL.
            data (dict[str, Any]): The request data (scope, client_id, etc.).

        Returns:
            dict[str, Any]: The device flow object containing device_code, user_code, etc.
        """

        try:
            device_flow_response = self.__app.http_client.post(
                device_auth_endpoint,
                data=data,
                headers={
                    "Accept": "application/json",
                    "Content-Type": "application/x-www-form-urlencoded;charset=UTF-8",
                },
            )
        except requests.exceptions.RequestException as e:
            raise CogniteAuthError("Error initiating device flow") from e

        # Handle MSAL NormalizedResponse which lacks .json() method
        if not hasattr(device_flow_response, "json"):
            try:
                return json.loads(device_flow_response.text)
            except (json.JSONDecodeError, TypeError, AttributeError) as e:
                raise CogniteAuthError("Unable to parse device flow response") from e

        # Normal response path
        try:
            return device_flow_response.json()
        except (json.JSONDecodeError, TypeError) as e:
            raise CogniteAuthError("Unable to parse device flow response") from e

    def _get_token(self, convert_timestamps: bool = True) -> dict[str, Any]:
        """Return a dictionary with the current token and expiry time."""
        if self._token_cache_path.exists():
            token = self._get_cached_token(self._token_cache_path)
        else:
            if _app := getattr(self, f"_{type(self).__name__}__app", None):
                if _app.token_cache.has_state_changed:
                    with open(self._token_cache_path, "w+") as fh:
                        fh.write(_app.token_cache.serialize())
            token = self._get_cached_token(self._token_cache_path)

        if convert_timestamps:
            if "AccessToken" in token:
                for key, value in token["AccessToken"].items():
                    for subkey in ["expires_on", "extended_expires_on", "cached_at"]:
                        if subkey in value:
                            value[subkey] = datetime.fromtimestamp(int(value[subkey])).isoformat()
        return token

    def _refresh_access_token(self) -> tuple[str, float]:
        # First check if a token cache exists on disk. If yes, find and use:
        # - A valid access token.
        # - A valid refresh token, and if so, use it automatically to redeem a new access token.
        credentials = None
        for token in self.__app.token_cache.search(self.__app.token_cache.CredentialType.REFRESH_TOKEN):
            if "expires_on" in token and token["expires_on"] > time.time():
                credentials = token
                break
        if credentials is not None:
            credentials = self.__app.client.obtain_token_by_refresh_token(credentials.get("secret", ""))
        else:
            for token in self.__app.token_cache.search(self.__app.token_cache.CredentialType.ACCESS_TOKEN):
                if expiry := int(token.get("expires_on", 0)) - time.time() > 0:
                    credentials = {
                        "access_token": token.get("secret"),
                        "expires_in": expiry,
                    }
                    break
        # If we're unable to find (or acquire a new) access token, we initiate the device code auth flow.
        # The msal device_code flow does not support setting the audience, so we need to handle it manually.
        # We use the http client instantiated as part of the msal client, as well as the details found
        # in oauth discovery.
        if credentials is None:
            data = {
                "scope": self.scope_string(),
                "client_id": self.client_id,
            }
            for key, value in self.__token_custom_args.items():
                data[key] = value

            device_flow_endpoint = self._get_device_authorization_endpoint()
            device_flow_response = self._get_device_code_response(device_flow_endpoint, data)
            if "verification_uri" in device_flow_response:
                print(  # noqa: T201
                    f"Visit {device_flow_response['verification_uri']} and enter the code: {device_flow_response.get('user_code', 'ERROR')}"
                )
            elif "message" in device_flow_response:
                print(  # noqa: T201
                    f"Device code: {device_flow_response.get('message', device_flow_response.get('user_code', 'ERROR'))}"
                )
            else:
                raise CogniteOAuthError(
                    device_flow_response.get("error", ""), device_flow_response.get("error_description", "")
                )

            if "interval" not in device_flow_response:
                # Set default interval according to standard
                device_flow_response["interval"] = 5
            if "expires_in" in device_flow_response:
                # msal library uses expires_at instead of the standard expires_in
                device_flow_response["expires_at"] = float(device_flow_response["expires_in"]) + time.time()
            # Poll for token
            credentials = self.__app.client.obtain_token_by_device_flow(
                flow=device_flow_response,
                data=dict(
                    data,
                    code=device_flow_response.get(
                        "device_code"
                    ),  # Hack from msal library to get the code from the device flow, not standard
                ),
            )

        self._verify_credentials(credentials)
        self.__app.token_cache.add(
            dict(credentials, environment=self.__app.authority.instance),
        )
        return credentials["access_token"], time.time() + float(credentials["expires_in"])

    @classmethod
    def load(cls, config: dict[str, Any] | str) -> OAuthDeviceCode:
        """Load a OAuth device code credential provider object from a YAML/JSON string or dict.

        Args:
            config (dict[str, Any] | str): A dictionary or YAML/JSON string containing configuration values defined in the OAuthDeviceCode class.

        Returns:
            OAuthDeviceCode: Initialized OAuthDeviceCode credential provider.

        Examples:

            >>> from cognite.client.credentials import OAuthDeviceCode
            >>> config = {
            ...     "authority_url": "https://login.microsoftonline.com/xyz",
            ...     "client_id": "abcd",
            ...     "scopes": ["https://greenfield.cognitedata.com/.default"],
            ... }
            >>> credentials = OAuthDeviceCode.load(config)
        """
        loaded = load_resource_to_dict(config)
        token_cache_path = loaded.pop("token_cache_path", None)
        return cls(
            authority_url=loaded.pop("authority_url"),
            client_id=loaded.pop("client_id"),
            scopes=loaded.pop("scopes", None),
            cdf_cluster=loaded.pop("cdf_cluster", None),
            oauth_discovery_url=loaded.pop("oauth_discovery_url", None),
            token_cache_path=Path(token_cache_path) if token_cache_path is not None else None,
            token_expiry_leeway_seconds=int(
                loaded.pop("token_expiry_leeway_seconds", _TOKEN_EXPIRY_LEEWAY_SECONDS_DEFAULT)
            ),
            clear_cache=loaded.pop("clear_cache", False),
            mem_cache_only=loaded.pop("mem_cache_only", False),
            **loaded,
        )

    @classmethod
    def default_for_entra_id(
        cls,
        tenant_id: str,
        client_id: str,
        cdf_cluster: str,
        token_cache_path: Path | None = None,
        token_expiry_leeway_seconds: int = _TOKEN_EXPIRY_LEEWAY_SECONDS_DEFAULT,
        clear_cache: bool = False,
        mem_cache_only: bool = False,
    ) -> OAuthDeviceCode:
        """
        Create an OAuthDeviceCode instance for Azure with default URLs and scopes. It uses the pre-configured Cognite
        app registration for device code flow. If you need device code flow with another app registration, instantiate
        OAuthDeviceCode directly.

        The default configuration creates the URLs based on the tenant id and cluster:

        * Authority URL: "https://login.microsoftonline.com/{tenant_id}"
        * Scopes: [f"https://{cdf_cluster}.cognitedata.com/.default"]

        Args:
            tenant_id (str): The Azure tenant id
            client_id (str): An app registration that allows device code flow.
            cdf_cluster (str): The CDF cluster where the CDF project is located.
            token_cache_path (Path | None): Location to store token cache, defaults to os temp directory/cognitetokencache.{client_id}.bin.
            token_expiry_leeway_seconds (int): The token is refreshed at the earliest when this number of seconds is left before expiry. Default: 30 sec
            clear_cache (bool): If True, the token cache will be cleared on initialization. Default: False
            mem_cache_only (bool): If True, the token cache will only be stored in memory. Default: False
        Returns:
            OAuthDeviceCode: An OAuthDeviceCode instance
        """
        return cls(
            authority_url=f"https://login.microsoftonline.com/{tenant_id}",
            client_id=client_id,  # Default application for CDF API for device code flow
            scopes=[
                f"https://{cdf_cluster}.cognitedata.com/IDENTITY",
                f"https://{cdf_cluster}.cognitedata.com/user_impersonation",
                "profile",
                "openid",
            ],
            token_cache_path=token_cache_path,
            token_expiry_leeway_seconds=token_expiry_leeway_seconds,
            clear_cache=clear_cache,
            mem_cache_only=mem_cache_only,
            audience=f"https://{cdf_cluster}.cognitedata.com",
        )

    @classmethod
    def default_for_azure_ad(cls, *args: Any, **kwargs: Any) -> OAuthDeviceCode:
        """Alias for :meth:`.OAuthDeviceCode.default_for_entra_id`.

        .. deprecated:: 8.x
            Use :meth:`.OAuthDeviceCode.default_for_entra_id` instead.

        """
        warnings.warn(
            "default_for_azure_ad() is deprecated; use default_for_entra_id() instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return cls.default_for_entra_id(*args, **kwargs)


class OAuthInteractive(_OAuthCredentialProviderWithTokenRefresh, _WithMsalSerializableTokenCache):
    """OAuth credential provider for an interactive login flow.

    Make sure you have http://localhost:port in Redirect URI in App Registration as type "Mobile and desktop applications".

    Args:
        authority_url (str): OAuth authority url
        client_id (str): Your application's client id.
        scopes (list[str]): A list of scopes.
        redirect_port (int): Redirect port defaults to 53000.
        token_cache_path (Path | None): Location to store token cache, defaults to os temp directory/cognitetokencache.{client_id}.bin.
        token_expiry_leeway_seconds (int): The token is refreshed at the earliest when this number of seconds is left before expiry. Default: 30 sec

    Examples:

            >>> from cognite.client.credentials import OAuthInteractive
            >>> oauth_provider = OAuthInteractive(
            ...     authority_url="https://login.microsoftonline.com/xyz",
            ...     client_id="abcd",
            ...     scopes=["https://greenfield.cognitedata.com/.default"],
            ... )
    """

    def __init__(
        self,
        authority_url: str,
        client_id: str,
        scopes: list[str],
        redirect_port: int = 53000,
        token_cache_path: Path | None = None,
        token_expiry_leeway_seconds: int = _TOKEN_EXPIRY_LEEWAY_SECONDS_DEFAULT,
    ) -> None:
        super().__init__(token_expiry_leeway_seconds)
        self.__authority_url = authority_url
        self.__client_id = client_id
        self.__scopes = scopes
        self.__redirect_port = redirect_port

        self._token_cache_path = self._resolve_token_cache_path(token_cache_path, client_id)
        self.__app = self._create_client_app(self._token_cache_path, client_id, authority_url)

    def __getstate__(self) -> dict[str, Any]:
        # PublicClientApplication is not picklable, temporarily remove:
        app_tmp, self.__app = self.__app, None
        state = super().__getstate__()
        self.__app = app_tmp
        return state

    def __setstate__(self, state: dict[str, Any]) -> None:
        super().__setstate__(state)
        self.__app = self._create_client_app(self._token_cache_path, self.__client_id, self.__authority_url)

    @property
    def authority_url(self) -> str:
        return self.__authority_url

    @property
    def client_id(self) -> str:
        return self.__client_id

    @property
    def scopes(self) -> list[str]:
        return self.__scopes

    def _refresh_access_token(self) -> tuple[str, float]:
        # First check if a token cache exists on disk. If yes, find and use:
        # - A valid access token.
        # - A valid refresh token, and if so, use it automatically to redeem a new access token.
        credentials = None
        if accounts := self.__app.get_accounts():
            credentials = self.__app.acquire_token_silent(scopes=self.__scopes, account=accounts[0])

        # If we're unable to find (or acquire a new) access token, we initiate the interactive auth flow:
        if credentials is None:
            credentials = self.__app.acquire_token_interactive(scopes=self.__scopes, port=self.__redirect_port)

        self._verify_credentials(credentials)
        return credentials["access_token"], time.time() + float(credentials["expires_in"])

    @classmethod
    def load(cls, config: dict[str, Any] | str) -> OAuthInteractive:
        """Load a OAuth interactive credential provider object from a YAML/JSON string or dict.

        Args:
            config (dict[str, Any] | str): A dictionary or YAML/JSON string containing configuration values defined in the OAuthInteractive class.

        Returns:
            OAuthInteractive: Initialized OAuthInteractive credential provider.

        Examples:

            >>> from cognite.client.credentials import OAuthInteractive
            >>> config = {
            ...     "authority_url": "https://login.microsoftonline.com/xyz",
            ...     "client_id": "abcd",
            ...     "scopes": ["https://greenfield.cognitedata.com/.default"],
            ... }
            >>> credentials = OAuthInteractive.load(config)
        """
        loaded = load_resource_to_dict(config)
        token_cache_path = loaded.get("token_cache_path")
        return cls(
            authority_url=loaded["authority_url"],
            client_id=loaded["client_id"],
            scopes=loaded["scopes"],
            redirect_port=int(loaded.get("redirect_port", 53000)),
            token_cache_path=Path(token_cache_path) if token_cache_path else None,
            token_expiry_leeway_seconds=int(
                loaded.get("token_expiry_leeway_seconds", _TOKEN_EXPIRY_LEEWAY_SECONDS_DEFAULT)
            ),
        )

    @classmethod
    def default_for_entra_id(
        cls,
        tenant_id: str,
        client_id: str,
        cdf_cluster: str,
        token_expiry_leeway_seconds: int = _TOKEN_EXPIRY_LEEWAY_SECONDS_DEFAULT,
        **token_custom_args: Any,
    ) -> OAuthInteractive:
        """
        Create an OAuthClientCredentials instance for Azure with default token URL and scopes.

        The default configuration creates the URLs based on the tenant id and cluster:

        * Authority URL: "https://login.microsoftonline.com/{tenant_id}"
        * Scopes: [f"https://{cdf_cluster}.cognitedata.com/.default"]

        Args:
            tenant_id (str): The Azure tenant id
            client_id (str): Your application's client id.
            cdf_cluster (str): The CDF cluster where the CDF project is located.
            token_expiry_leeway_seconds (int): The token is refreshed at the earliest when this number of seconds is left before expiry. Default: 30 sec
            **token_custom_args (Any): Optional additional arguments to pass as query parameters to the token fetch request.

        Returns:
            OAuthInteractive: An OAuthInteractive instance
        """
        return cls(
            authority_url=f"https://login.microsoftonline.com/{tenant_id}",
            client_id=client_id,
            scopes=[f"https://{cdf_cluster}.cognitedata.com/.default"],
            token_expiry_leeway_seconds=token_expiry_leeway_seconds,
            **token_custom_args,
        )

    @classmethod
    def default_for_azure_ad(cls, *args: Any, **kwargs: Any) -> OAuthInteractive:
        """Alias for :meth:`.OAuthInteractive.default_for_entra_id`.

        .. deprecated:: 8.x
            Use :meth:`.OAuthInteractive.default_for_entra_id` instead.

        """
        warnings.warn(
            "default_for_azure_ad() is deprecated; use default_for_entra_id() instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return cls.default_for_entra_id(*args, **kwargs)


class OAuthClientCredentials(_OAuthCredentialProviderWithTokenRefresh):
    """OAuth credential provider for the "Client Credentials" flow.

    Args:
        token_url (str): OAuth token url
        client_id (str): Your application's client id.
        client_secret (str): Your application's client secret
        scopes (list[str] | None): A list of scopes.
        token_expiry_leeway_seconds (int): The token is refreshed at the earliest when this number of seconds is left before expiry. Default: 30 sec
        **token_custom_args (Any): Optional additional arguments to pass as query parameters to the token fetch request.

    Examples:

            >>> from cognite.client.credentials import OAuthClientCredentials
            >>> import os
            >>> oauth_provider = OAuthClientCredentials(
            ...     token_url="https://login.microsoftonline.com/xyz/oauth2/v2.0/token",
            ...     client_id="abcd",
            ...     client_secret=os.environ["OAUTH_CLIENT_SECRET"],
            ...     scopes=["https://greenfield.cognitedata.com/.default"],
            ...     # Any additional IDP-specific token args. e.g.
            ...     audience="some-audience"
            ... )
    """

    def __init__(
        self,
        token_url: str,
        client_id: str,
        client_secret: str,
        scopes: list[str] | None,
        token_expiry_leeway_seconds: int = _TOKEN_EXPIRY_LEEWAY_SECONDS_DEFAULT,
        **token_custom_args: Any,
    ) -> None:
        super().__init__(token_expiry_leeway_seconds)
        self.__token_url = token_url
        self.__client_id = client_id
        self.__client_secret = client_secret
        self.__scopes = scopes
        self.__token_custom_args: dict[str, Any] = token_custom_args
        self.__oauth = self._create_oauth_client()
        self._validate_token_custom_args()

    def _create_oauth_client(self) -> OAuth2Client:
        from cognite.client.config import global_config

        return OAuth2Client(client_id=self.__client_id, scope=self.__scopes, verify=not global_config.disable_ssl)

    def _validate_token_custom_args(self) -> None:
        # We make sure that whatever is passed as part of 'token_custom_args' can't set or override any of the
        # named parameters that 'fetch_token' accepts:
        reserved = set(inspect.signature(self.__oauth.fetch_token).parameters) - {"kwargs"}
        if bad_args := reserved.intersection(self.__token_custom_args):
            raise TypeError(
                f"The following reserved token custom arg(s) were passed: {sorted(bad_args)}. The full list of "
                f"reserved custom args is: {sorted(reserved)}."
            )

    def __getstate__(self) -> dict[str, Any]:
        # OAuth2Client is not picklable, temporarily remove:
        oauth_session_tmp, self.__oauth = self.__oauth, None
        state = super().__getstate__()
        self.__oauth = oauth_session_tmp
        return state

    def __setstate__(self, state: dict[str, Any]) -> None:
        super().__setstate__(state)
        self.__oauth = self._create_oauth_client()

    @property
    def token_url(self) -> str:
        return self.__token_url

    @property
    def client_id(self) -> str:
        return self.__client_id

    @property
    def client_secret(self) -> str:
        return self.__client_secret

    @property
    def scopes(self) -> list[str] | None:
        return self.__scopes

    @property
    def token_custom_args(self) -> dict[str, Any]:
        return self.__token_custom_args

    def _refresh_access_token(self) -> tuple[str, float]:
        try:
            credentials = self.__oauth.fetch_token(
                url=self.__token_url, client_secret=self.__client_secret, **self.__token_custom_args
            )
            # Azure gives 'expires_at' directly, but it's not a part of the RFC:
            return credentials["access_token"], time.time() + float(credentials["expires_in"])

        except OAuthError as err:
            raise CogniteOAuthError(err.error, err.description) from err

    @classmethod
    def load(cls, config: dict[str, Any] | str) -> OAuthClientCredentials:
        """Load a OAuth client credentials credential provider object from a YAML/JSON string or dict.

        Args:
            config (dict[str, Any] | str): A dictionary or YAML/JSON string containing configuration values defined in the OAuthClientCredentials class.

        Returns:
            OAuthClientCredentials: Initialized OAuthClientCredentials credential provider.

        Examples:

            >>> from cognite.client.credentials import OAuthClientCredentials
            >>> import os
            >>> config = {
            ...     "token_url": "https://login.microsoftonline.com/xyz/oauth2/v2.0/token",
            ...     "client_id": "abcd",
            ...     "client_secret": os.environ["OAUTH_CLIENT_SECRET"],
            ...     "scopes": ["https://greenfield.cognitedata.com/.default"],
            ...     "audience": "some-audience"
            ... }
            >>> credentials = OAuthClientCredentials.load(config)
        """
        loaded = load_resource_to_dict(config).copy()  # doing a shallow copy to avoid mutating the user input config
        return cls(
            token_url=loaded.pop("token_url"),
            client_id=loaded.pop("client_id"),
            client_secret=loaded.pop("client_secret"),
            scopes=loaded.pop("scopes"),
            token_expiry_leeway_seconds=int(
                loaded.pop("token_expiry_leeway_seconds", _TOKEN_EXPIRY_LEEWAY_SECONDS_DEFAULT)
            ),
            **loaded,
        )

    @classmethod
    def default_for_entra_id(
        cls,
        tenant_id: str,
        client_id: str,
        client_secret: str,
        cdf_cluster: str,
        token_expiry_leeway_seconds: int = _TOKEN_EXPIRY_LEEWAY_SECONDS_DEFAULT,
        **token_custom_args: Any,
    ) -> OAuthClientCredentials:
        """
        Create an OAuthClientCredentials instance for Azure with default token URL and scopes.

        The default configuration creates the URLs based on the tenant id and cluster/oauth2/v2.0/token:

        * Token URL: "https://login.microsoftonline.com/{tenant_id}"
        * Scopes: [f"https://{cdf_cluster}.cognitedata.com/.default"]

        Args:
            tenant_id (str): The Azure tenant id
            client_id (str): Your application's client id.
            client_secret (str): Your application's client secret.
            cdf_cluster (str): The CDF cluster where the CDF project is located.
            token_expiry_leeway_seconds (int): The token is refreshed at the earliest when this number of seconds is left before expiry. Default: 30 sec
            **token_custom_args (Any): Optional additional arguments to pass as query parameters to the token fetch request.

        Returns:
            OAuthClientCredentials: An OAuthClientCredentials instance
        """
        return cls(
            token_url=f"https://login.microsoftonline.com/{tenant_id}/oauth2/v2.0/token",
            client_id=client_id,
            client_secret=client_secret,
            scopes=[f"https://{cdf_cluster}.cognitedata.com/.default"],
            token_expiry_leeway_seconds=token_expiry_leeway_seconds,
            **token_custom_args,
        )

    @classmethod
    def default_for_azure_ad(cls, *args: Any, **kwargs: Any) -> OAuthClientCredentials:
        """Alias for :meth:`.OAuthClientCredentials.default_for_entra_id`.

        .. deprecated:: 8.x
            Use :meth:`.OAuthClientCredentials.default_for_entra_id` instead.

        """
        warnings.warn(
            "default_for_azure_ad() is deprecated; use default_for_entra_id() instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return cls.default_for_entra_id(*args, **kwargs)


class OAuthClientCertificate(_OAuthCredentialProviderWithTokenRefresh):
    """OAuth credential provider for authenticating with a client certificate.

    Args:
        authority_url (str): OAuth authority url
        client_id (str): Your application's client id.
        cert_thumbprint (str): Your certificate's thumbprint. You get it when you upload your certificate to Azure AD.
        certificate (str): Your private certificate, typically read from a .pem file
        scopes (list[str]): A list of scopes.
        token_expiry_leeway_seconds (int): The token is refreshed at the earliest when this number of seconds is left before expiry. Default: 30 sec

    Examples:

            >>> from cognite.client.credentials import OAuthClientCertificate
            >>> from pathlib import Path
            >>> oauth_provider = OAuthClientCertificate(
            ...     authority_url="https://login.microsoftonline.com/xyz",
            ...     client_id="abcd",
            ...     cert_thumbprint="XYZ123",
            ...     certificate=Path("certificate.pem").read_text(),
            ...     scopes=["https://greenfield.cognitedata.com/.default"],
            ... )
    """

    def __init__(
        self,
        authority_url: str,
        client_id: str,
        cert_thumbprint: str,
        certificate: str,
        scopes: list[str],
        token_expiry_leeway_seconds: int = _TOKEN_EXPIRY_LEEWAY_SECONDS_DEFAULT,
    ) -> None:
        super().__init__(token_expiry_leeway_seconds)
        self.__authority_url = authority_url
        self.__client_id = client_id
        self.__cert_thumbprint = cert_thumbprint
        self.__certificate = certificate
        self.__scopes = scopes

        self.__app = ConfidentialClientApplication(
            client_id=self.__client_id,
            authority=self.__authority_url,
            client_credential={"thumbprint": self.__cert_thumbprint, "private_key": self.__certificate},
        )

    @property
    def authority_url(self) -> str:
        return self.__authority_url

    @property
    def client_id(self) -> str:
        return self.__client_id

    @property
    def cert_thumbprint(self) -> str:
        return self.__cert_thumbprint

    @property
    def certificate(self) -> str:
        return self.__certificate

    @property
    def scopes(self) -> list[str]:
        return self.__scopes

    def _refresh_access_token(self) -> tuple[str, float]:
        credentials = self.__app.acquire_token_for_client(scopes=self.__scopes)

        self._verify_credentials(credentials)
        return credentials["access_token"], time.time() + float(credentials["expires_in"])

    @classmethod
    def load(cls, config: dict[str, Any] | str) -> OAuthClientCertificate:
        """Load a OAuth client certificate credential provider object from a YAML/JSON string or dict.

        Args:
            config (dict[str, Any] | str): A dictionary or YAML/JSON string containing configuration values defined in the OAuthClientCertificate class.

        Returns:
            OAuthClientCertificate: Initialized OAuthClientCertificate credential provider.

        Examples:

            >>> from cognite.client.credentials import OAuthClientCertificate
            >>> from pathlib import Path
            >>> config = {
            ...     "authority_url": "https://login.microsoftonline.com/xyz",
            ...     "client_id": "abcd",
            ...     "cert_thumbprint": "XYZ123",
            ...     "certificate": Path("certificate.pem").read_text(),
            ...     "scopes": ["https://greenfield.cognitedata.com/.default"],
            ... }
            >>> credentials = OAuthClientCertificate.load(config)
        """
        loaded = load_resource_to_dict(config)
        return cls(
            authority_url=loaded["authority_url"],
            client_id=loaded["client_id"],
            cert_thumbprint=loaded["cert_thumbprint"],
            certificate=loaded["certificate"],
            scopes=loaded["scopes"],
            token_expiry_leeway_seconds=int(
                loaded.get("token_expiry_leeway_seconds", _TOKEN_EXPIRY_LEEWAY_SECONDS_DEFAULT)
            ),
        )


_SUPPORTED_CREDENTIAL_TYPES = MappingProxyType(
    {
        "token": Token,
        "client_credentials": OAuthClientCredentials,
        "interactive": OAuthInteractive,
        "device_code": OAuthDeviceCode,
        "client_certificate": OAuthClientCertificate,
    }
)



"""
===============================================================================
7987a420577b0616995be35641dc2168
This file is auto-generated from the Async API modules, - do not edit manually!
===============================================================================
"""

from __future__ import annotations

from collections.abc import Iterator, Sequence
from typing import TYPE_CHECKING, overload

from cognite.client import AsyncCogniteClient
from cognite.client._constants import DEFAULT_LIMIT_READ
from cognite.client._sync_api_client import SyncAPIClient
from cognite.client.data_classes.data_modeling.spaces import Space, SpaceApply, SpaceList
from cognite.client.utils._async_helpers import SyncIterator, run_sync
from cognite.client.utils.useful_types import SequenceNotStr

if TYPE_CHECKING:
    from cognite.client import AsyncCogniteClient


class SyncSpacesAPI(SyncAPIClient):
    """Auto-generated, do not modify manually."""

    def __init__(self, async_client: AsyncCogniteClient) -> None:
        self.__async_client = async_client

    @overload
    def __call__(self, chunk_size: None = None, limit: int | None = None) -> Iterator[Space]: ...

    @overload
    def __call__(self, chunk_size: int, limit: int | None = None) -> Iterator[SpaceList]: ...

    def __call__(
        self, chunk_size: int | None = None, limit: int | None = None
    ) -> Iterator[Space] | Iterator[SpaceList]:
        """
        Iterate over spaces

        Fetches spaces as they are iterated over, so you keep a limited number of spaces in memory.

        Args:
            chunk_size (int | None): Number of spaces to return in each chunk. Defaults to yielding one space a time.
            limit (int | None): Maximum number of spaces to return. Defaults to returning all items.

        Yields:
            Space | SpaceList: yields Space one by one if chunk_size is not specified, else SpaceList objects.
        """  # noqa: DOC404
        yield from SyncIterator(self.__async_client.data_modeling.spaces(chunk_size=chunk_size, limit=limit))  # type: ignore [misc]

    @overload
    def retrieve(self, spaces: str) -> Space | None: ...

    @overload
    def retrieve(self, spaces: SequenceNotStr[str]) -> SpaceList: ...

    def retrieve(self, spaces: str | SequenceNotStr[str]) -> Space | SpaceList | None:
        """
        `Retrieve one or more spaces. <https://developer.cognite.com/api#tag/Spaces/operation/bySpaceIdsSpaces>`_

        Args:
            spaces (str | SequenceNotStr[str]): Space ID

        Returns:
            Space | SpaceList | None: Requested space or None if it does not exist.

        Examples:

                >>> from cognite.client import CogniteClient, AsyncCogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> res = client.data_modeling.spaces.retrieve(spaces='mySpace')

            Get multiple spaces by id:

                >>> res = client.data_modeling.spaces.retrieve(spaces=["MySpace", "MyAwesomeSpace", "MyOtherSpace"])
        """
        return run_sync(self.__async_client.data_modeling.spaces.retrieve(spaces=spaces))

    def delete(self, spaces: str | SequenceNotStr[str]) -> list[str]:
        """
        `Delete one or more spaces <https://developer.cognite.com/api#tag/Spaces/operation/deleteSpacesV3>`_

        Args:
            spaces (str | SequenceNotStr[str]): ID or ID list ids of spaces.
        Returns:
            list[str]: The space(s) which has been deleted.
        Examples:

            Delete spaces by id:

                >>> from cognite.client import CogniteClient, AsyncCogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> client.data_modeling.spaces.delete(spaces=["mySpace", "myOtherSpace"])
        """
        return run_sync(self.__async_client.data_modeling.spaces.delete(spaces=spaces))

    def list(self, limit: int | None = DEFAULT_LIMIT_READ, include_global: bool = False) -> SpaceList:
        """
        `List spaces <https://developer.cognite.com/api#tag/Spaces/operation/listSpacesV3>`_

        Args:
            limit (int | None): Maximum number of spaces to return. Defaults to 10. Set to -1, float("inf") or None to return all items.
            include_global (bool): Whether to include global spaces. Defaults to False.

        Returns:
            SpaceList: List of requested spaces

        Examples:

            List spaces and filter on max start time:

                >>> from cognite.client import CogniteClient, AsyncCogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> space_list = client.data_modeling.spaces.list(limit=5)

            Iterate over spaces, one-by-one:

                >>> for space in client.data_modeling.spaces():
                ...     space  # do something with the space

            Iterate over chunks of spaces to reduce memory load:

                >>> for space_list in client.data_modeling.spaces(chunk_size=2500):
                ...     space_list # do something with the spaces
        """
        return run_sync(self.__async_client.data_modeling.spaces.list(limit=limit, include_global=include_global))

    @overload
    def apply(self, spaces: Sequence[SpaceApply]) -> SpaceList: ...

    @overload
    def apply(self, spaces: SpaceApply) -> Space: ...

    def apply(self, spaces: SpaceApply | Sequence[SpaceApply]) -> Space | SpaceList:
        """
        `Create or patch one or more spaces. <https://developer.cognite.com/api#tag/Spaces/operation/ApplySpaces>`_

        Args:
            spaces (SpaceApply | Sequence[SpaceApply]): Space | Sequence[Space]): Space or spaces of spacesda to create or update.

        Returns:
            Space | SpaceList: Created space(s)

        Examples:

            Create new spaces:

                >>> from cognite.client import CogniteClient
                >>> from cognite.client.data_classes.data_modeling import SpaceApply
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> spaces = [SpaceApply(space="mySpace", description="My first space", name="My Space"),
                ... SpaceApply(space="myOtherSpace", description="My second space", name="My Other Space")]
                >>> res = client.data_modeling.spaces.apply(spaces)
        """
        return run_sync(self.__async_client.data_modeling.spaces.apply(spaces=spaces))



from __future__ import annotations

import asyncio
import threading
import warnings
from collections import UserList
from collections.abc import Callable, Coroutine
from functools import cache
from typing import (
    Any,
    NoReturn,
    TypeVar,
    cast,
)

from cognite.client._constants import _RUNNING_IN_BROWSER
from cognite.client.exceptions import CogniteAPIError, CogniteDuplicatedError, CogniteImportError, CogniteNotFoundError
from cognite.client.utils._auxiliary import no_op


@cache
def get_global_semaphore() -> asyncio.BoundedSemaphore:
    from cognite.client import global_config

    return asyncio.BoundedSemaphore(global_config.max_workers)


@cache
def get_global_datapoints_semaphore() -> asyncio.BoundedSemaphore:
    from cognite.client import global_config

    return asyncio.BoundedSemaphore(global_config.max_workers)


@cache
def get_global_data_modeling_semaphore() -> asyncio.BoundedSemaphore:
    return asyncio.BoundedSemaphore(2)


class AsyncSDKTask:
    """
    This class stores info about a task that should be run asynchronously (like what function to call,
    and with what arguments). This is quite useful when we later need to map e.g. which input arguments
    (typically, which identifiers) resulted in some failed API request.

    It has a special ``__getitem__`` method for easy argument access:

        >>> task = AsyncSDKTask(some_fn, 123, payload={"a": 1, "b": 2})
        >>> task[0]  # gets the first positional argument
        123
        >>> task["payload"]  # gets the "payload" keyword argument
        {"a": 1, "b": 2}
    """

    def __init__(self, fn: Callable[..., Coroutine], /, *args: Any, **kwargs: Any) -> None:
        self.fn = fn
        self._args = args
        self._kwargs = kwargs
        self._async_task: asyncio.Task | None = None

    def __getitem__(self, item: int | str) -> Any:
        match item:
            case int():
                return self._args[item]
            case str():
                return self._kwargs[item]
            case _:
                raise TypeError("Use int to get 'args' and str to get 'kwargs', e.g. task[0], task['payload']")

    def schedule(self) -> asyncio.Task:
        """Schedule the task for execution. Can only be called once."""
        if self._async_task is not None:
            raise RuntimeError("Task has already been scheduled")

        self._async_task = asyncio.create_task(self.fn(*self._args, **self._kwargs))
        return self._async_task

    @property
    def async_task(self) -> asyncio.Task:
        if self._async_task is None:
            raise RuntimeError("Task has not been scheduled yet")
        return self._async_task

    def exception(self) -> BaseException | None:
        return self.async_task.exception()

    def result(self) -> Any:
        return self.async_task.result()

    def cancelled(self) -> bool:
        return self.async_task.cancelled()


class TasksSummary:
    def __init__(
        self,
        results: list[Any],
        successful_tasks: list[AsyncSDKTask],
        unsuccessful_tasks: list[AsyncSDKTask] | None = None,
        skipped_tasks: list[AsyncSDKTask] | None = None,
        exceptions: list[BaseException] | None = None,
    ) -> None:
        self.results = results
        self.successful_tasks = successful_tasks
        self.unknown_tasks, self.failed_tasks = self._categorize_failed_vs_unknown(unsuccessful_tasks or [])
        self.skipped_tasks = skipped_tasks or []

        self.not_found_error: CogniteNotFoundError | None = None
        self.duplicated_error: CogniteDuplicatedError | None = None
        self.unknown_error: BaseException | None = None
        self.missing, self.duplicated, self.cluster, self.project = self._inspect_exceptions(exceptions or [])

    @staticmethod
    def _categorize_failed_vs_unknown(
        unsuccessful_tasks: list[AsyncSDKTask],
    ) -> tuple[list[AsyncSDKTask], list[AsyncSDKTask]]:
        from cognite.client._basic_api_client import FailedRequestHandler

        unknown_and_failed: tuple[list[AsyncSDKTask], list[AsyncSDKTask]] = [], []
        for task in unsuccessful_tasks:
            err = cast(BaseException, task.exception())  # Task is unsuccessful exactly because this is set
            is_failed = FailedRequestHandler.classify_error(err) == "failed"
            unknown_and_failed[is_failed].append(task)
        return unknown_and_failed

    def joined_results(self, unwrap_fn: Callable = no_op) -> list:
        joined_results: list = []
        for result in self.results:
            unwrapped = unwrap_fn(result)
            if isinstance(unwrapped, (list, UserList)):
                joined_results.extend(unwrapped)
            else:
                joined_results.append(unwrapped)
        return joined_results

    @property
    def raw_api_responses(self) -> list[dict[str, Any]]:
        return [res.json() for res in self.results]

    def raise_compound_exception_if_failed_tasks(
        self, task_unwrap_fn: Callable = no_op, task_list_element_unwrap_fn: Callable | None = None
    ) -> None:
        if not (self.unknown_error or self.not_found_error or self.duplicated_error):
            return None

        if task_list_element_unwrap_fn is None:
            successful = [task_unwrap_fn(t) for t in self.successful_tasks]
            unknown = [task_unwrap_fn(t) for t in self.unknown_tasks]
            failed = [task_unwrap_fn(t) for t in self.failed_tasks]
            skipped = [task_unwrap_fn(t) for t in self.skipped_tasks]
        else:
            successful = [task_list_element_unwrap_fn(el) for t in self.successful_tasks for el in task_unwrap_fn(t)]
            unknown = [task_list_element_unwrap_fn(el) for t in self.unknown_tasks for el in task_unwrap_fn(t)]
            failed = [task_list_element_unwrap_fn(el) for t in self.failed_tasks for el in task_unwrap_fn(t)]
            skipped = [task_list_element_unwrap_fn(el) for t in self.skipped_tasks for el in task_unwrap_fn(t)]

        if self.unknown_error:
            self._raise_basic_api_error(successful=successful, failed=failed, unknown=unknown, skipped=skipped)

        if self.not_found_error:
            self._raise_specific_error(
                cause=self.not_found_error,
                error=CogniteNotFoundError,
                successful=successful,
                failed=failed,
                unknown=unknown,
                skipped=skipped,
            )
        if self.duplicated_error:
            self._raise_specific_error(
                cause=self.duplicated_error,
                error=CogniteDuplicatedError,
                successful=successful,
                failed=failed,
                unknown=unknown,
                skipped=skipped,
            )

    def _inspect_exceptions(self, exceptions: list[BaseException]) -> tuple[list, list, str | None, str | None]:
        cluster = None
        project = None
        missing: list[dict] = []
        duplicated: list[dict] = []
        for exc in exceptions:
            match exc:
                case CogniteNotFoundError():
                    missing.extend(exc.missing)
                    self.not_found_error = exc
                case CogniteDuplicatedError():
                    duplicated.extend(exc.duplicated)
                    self.duplicated_error = exc
                case CogniteAPIError():
                    self.unknown_error = exc
                case _:
                    self.unknown_error = exc
                    continue

            cluster = cluster or exc.cluster
            project = project or exc.project

        return missing, duplicated, cluster, project

    def _raise_basic_api_error(self, successful: list, failed: list, unknown: list, skipped: list) -> NoReturn:
        if isinstance(self.unknown_error, CogniteAPIError) and (failed or unknown):
            raise CogniteAPIError(
                message=self.unknown_error.message,
                code=self.unknown_error.code,
                x_request_id=self.unknown_error.x_request_id,
                missing=self.missing,
                duplicated=self.duplicated,
                extra=self.unknown_error.extra,
                cluster=self.cluster,
                project=self.project,
                successful=successful,
                failed=failed,
                unknown=unknown,
                skipped=skipped,
            )
        raise self.unknown_error  # type: ignore [misc]

    def _raise_specific_error(
        self,
        cause: CogniteAPIError,
        error: type[CogniteNotFoundError | CogniteDuplicatedError],
        successful: list,
        failed: list,
        unknown: list,
        skipped: list,
    ) -> NoReturn:
        raise error(
            message=cause.message,
            code=cause.code,
            x_request_id=cause.x_request_id,
            missing=self.missing,
            duplicated=self.duplicated,
            extra=cause.extra,
            cluster=self.cluster,
            project=self.project,
            successful=successful,
            failed=failed,
            unknown=unknown,
            skipped=skipped,
        ) from cause


_T = TypeVar("_T")


class EventLoopThreadExecutor(threading.Thread):
    def __init__(self, loop: asyncio.AbstractEventLoop | None = None, daemon: bool = True) -> None:
        super().__init__(name=type(self).__name__, daemon=daemon)
        self._inside_jupyter = self._detect_jupyter()
        self._event_loop = self._patch_loop_for_jupyter(loop) or asyncio.new_event_loop()

    @staticmethod
    def _detect_jupyter() -> bool:
        try:
            from IPython import get_ipython  # type: ignore [attr-defined]

            return "IPKernelApp" in get_ipython().config
        except Exception:
            return False

    def _patch_loop_for_jupyter(self, loop: asyncio.AbstractEventLoop | None) -> asyncio.AbstractEventLoop | None:
        """
        From the 'nest_asyncio' package: By design asyncio does not allow its event loop to be nested. This presents a
        practical problem: When in an environment where the event loop is already running it's impossible to run
        tasks and wait for the result.
        """
        if not self._inside_jupyter:
            return loop

        if loop is None:
            try:
                import nest_asyncio  # type: ignore [import-not-found]
            except ImportError:
                raise CogniteImportError(
                    module="nest_asyncio",
                    message="Inside Jupyter notebooks, the 'nest_asyncio' package is required if you want to use the "
                    '"non-async" CogniteClient. This is because Jupyter already runs an event loop that we need '
                    "to patch (`pip install nest_asyncio`). Alternatively, you can use the AsyncCogniteClient "
                    "which does not require any extra packages, but requires the use of 'await', e.g.: "
                    "`dps = await async_client.time_series.data.retrieve(...)`",
                ) from None
            try:
                # Jupyter: reuse the already running loop but patch it:
                loop = asyncio.get_running_loop()
                nest_asyncio.apply(loop)
                return loop
            except RuntimeError:
                return None  # this would be very unexpected
        else:
            warnings.warn(
                RuntimeWarning(
                    "Overriding the event loop is not recommended inside Jupyter notebooks "
                    "since Jupyter already runs an event loop. Proceeding with the provided loop anyway, "
                    "beware of potential issues."
                )
            )
            return loop

    def run(self) -> None:
        if not self._inside_jupyter:
            asyncio.set_event_loop(self._event_loop)
            self._event_loop.run_forever()

    def stop(self) -> None:
        if not self._inside_jupyter:
            self._event_loop.call_soon_threadsafe(self._event_loop.stop)
            self.join()

    def run_coro(self, coro: Coroutine[_T, Any, _T], timeout: float | None = None) -> _T:
        if self._inside_jupyter:
            return asyncio.get_event_loop().run_until_complete(coro)
        else:
            return asyncio.run_coroutine_threadsafe(coro, self._event_loop).result(timeout)


class _PyodideEventLoopExecutor:
    def __init__(self, loop: asyncio.AbstractEventLoop | None = None) -> None:
        import pyodide  # type: ignore [import-not-found]

        if loop is not None:
            raise RuntimeError("Overriding the event loop is not possible in the browser")

        elif not pyodide.ffi.can_run_sync():
            warnings.warn(
                RuntimeWarning(
                    "Browser most likely not supported, please use a Chromium-based browser like Chrome or Microsoft "
                    "Edge. Reason: WebAssembly stack switching is not supported in this JavaScript runtime. "
                    "Note: You can always use the AsyncCogniteClient, but it requires the use of 'await', e.g.: "
                    "`dps = await client.time_series.data.retrieve(...)`"
                )
            )
        self.run_coro = pyodide.ffi.run_sync

    def start(self) -> None:
        pass


# We need this in order to support a synchronous Cognite client.
_INTERNAL_EVENT_LOOP_THREAD_EXECUTOR_SINGLETON: EventLoopThreadExecutor


class ConcurrencySettings:
    @classmethod
    def _get_event_loop_executor(cls) -> EventLoopThreadExecutor:
        global _INTERNAL_EVENT_LOOP_THREAD_EXECUTOR_SINGLETON
        try:
            return _INTERNAL_EVENT_LOOP_THREAD_EXECUTOR_SINGLETON
        except NameError:
            # First time we need to initialize:
            from cognite.client import global_config

            ex_cls = EventLoopThreadExecutor
            if _RUNNING_IN_BROWSER:
                ex_cls = cast(type[EventLoopThreadExecutor], _PyodideEventLoopExecutor)

            executor = _INTERNAL_EVENT_LOOP_THREAD_EXECUTOR_SINGLETON = ex_cls(global_config.event_loop)
            executor.start()
            return executor


async def execute_async_tasks_with_fail_fast(tasks: list[AsyncSDKTask]) -> TasksSummary:
    # If no future raises an exception then this is equivalent to asyncio.ALL_COMPLETED:
    done, pending = await asyncio.wait(
        [task.schedule() for task in tasks],
        return_when=asyncio.FIRST_EXCEPTION,
    )
    if all(task.exception() is None for task in done):
        return TasksSummary([task.result() for task in done], successful_tasks=tasks)

    # Something failed, and because of fail-fast, we (attempt to) cancel all pending tasks:
    if pending:  # while we are waiting on 3.11 and asyncio.TaskGroup...
        for unfinished in pending:
            unfinished.cancel()

        # Wait for all cancellations to be processed:
        await asyncio.gather(*pending, return_exceptions=True)

    result, successful, unsuccessful, skipped, exceptions = [], [], [], [], []
    for task in tasks:
        if task.cancelled():
            skipped.append(task)
        elif err := task.exception():
            exceptions.append(err)
            unsuccessful.append(task)
        else:
            result.append(task.result())
            successful.append(task)

    return TasksSummary(
        result,
        successful_tasks=successful,
        unsuccessful_tasks=unsuccessful,
        skipped_tasks=skipped,
        exceptions=exceptions,
    )


async def execute_async_tasks(tasks: list[AsyncSDKTask], fail_fast: bool = False) -> TasksSummary:
    if not tasks:
        return TasksSummary([], successful_tasks=[], unsuccessful_tasks=[], exceptions=[])

    elif fail_fast:
        return await execute_async_tasks_with_fail_fast(tasks)

    await asyncio.wait([task.schedule() for task in tasks], return_when=asyncio.ALL_COMPLETED)

    results, successful, unsuccessful, exceptions = [], [], [], []
    for task in tasks:
        if err := task.exception():
            exceptions.append(err)
            unsuccessful.append(task)
        else:
            results.append(task.result())
            successful.append(task)

    return TasksSummary(
        results,
        successful_tasks=successful,
        unsuccessful_tasks=unsuccessful,
        exceptions=exceptions,
    )



from __future__ import annotations

from cognite.client._api_client import APIClient
from cognite.client.utils._url import interpolate_and_url_encode


class ThreeDFilesAPI(APIClient):
    _RESOURCE_PATH = "/3d/files"

    async def retrieve(self, id: int) -> bytes:
        """`Retrieve the contents of a 3d file by id. <https://developer.cognite.com/api#tag/3D-Files/operation/get3DFile>`_

        Args:
            id (int): The id of the file to retrieve.

        Returns:
            bytes: The contents of the file.

        Example:

            Retrieve the contents of a 3d file by id:

                >>> from cognite.client import CogniteClient, AsyncCogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> res = client.three_d.files.retrieve(1)
        """
        path = interpolate_and_url_encode(self._RESOURCE_PATH + "/{}", id)
        return (await self._get(path)).content



"""
===============================================================================
22ce5358beca6d957d6ed92d895118ae
This file is auto-generated from the Async API modules, - do not edit manually!
===============================================================================
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from cognite.client import AsyncCogniteClient
from cognite.client._sync_api.data_modeling.space_statistics import SyncSpaceStatisticsAPI
from cognite.client._sync_api_client import SyncAPIClient
from cognite.client.data_classes.data_modeling.statistics import ProjectStatistics
from cognite.client.utils._async_helpers import run_sync

if TYPE_CHECKING:
    from cognite.client import AsyncCogniteClient


class SyncStatisticsAPI(SyncAPIClient):
    """Auto-generated, do not modify manually."""

    def __init__(self, async_client: AsyncCogniteClient) -> None:
        self.__async_client = async_client
        self.spaces = SyncSpaceStatisticsAPI(async_client)

    def project(self) -> ProjectStatistics:
        """
        `Retrieve project-wide usage data and limits <https://developer.cognite.com/api#tag/Statistics/operation/getStatistics>`_

        Returns the usage data and limits for a project's data modelling usage, including data model schemas and graph instances

        Returns:
            ProjectStatistics: The requested statistics and limits

        Examples:

            Fetch project statistics (and limits) and check the current number of data models vs.
            and how many more can be created:

                >>> from cognite.client import CogniteClient, AsyncCogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> stats = client.data_modeling.statistics.project()
                >>> data_model_count = stats.data_models.count
                >>> available_count = stats.data_models.limit - data_model_count
        """
        return run_sync(self.__async_client.data_modeling.statistics.project())
